/*******************************************************************************
 *
 * Copyright (C) 2022 NETINT Technologies
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 ******************************************************************************/

/*!*****************************************************************************
*  \file   ni_dec_api_logan.c
*
*  \brief  NETINT decoder API file
*
*******************************************************************************/

#include "ni_dec_api_logan.h"
#define NI_LOGAN_DEC_PKT_BUF_NUM_MAX 300

void ni_logan_buf_pool_free(void *opaque, uint8_t *data)
{
  if (data)
  {
    ni_logan_buf_t *buf = (ni_logan_buf_t *) opaque;
    ni_logan_decoder_frame_buffer_pool_return_buf(buf, (ni_logan_buf_pool_t *)buf->pool);
  }
}
int ni_logan_xcoder_dec_close(ni_logan_dec_context_t *ni_dec_ctx)
{
  ni_logan_retcode_t ret;

  ni_logan_session_context_t *p_session_ctx = NULL;
  ni_log(NI_LOG_DEBUG, "ni_logan_xcoder_dec_close enter\n");
  if (!ni_dec_ctx)
  {
    ni_log(NI_LOG_ERROR, "ni_dec_ctx is null\n");
    return NI_LOGAN_RETCODE_INVALID_PARAM;
  }
  p_session_ctx = ni_dec_ctx->p_session_ctx;
  if (!p_session_ctx)
  {
    ni_log(NI_LOG_ERROR, "%s(): line %d, Invalid Parameter, %p\n",
           __FUNCTION__, __LINE__, p_session_ctx);
    return NI_LOGAN_RETCODE_INVALID_PARAM;
  }

  ret = ni_logan_device_session_close(p_session_ctx, ni_dec_ctx->eos, NI_LOGAN_DEVICE_TYPE_DECODER);
  if (NI_LOGAN_RETCODE_SUCCESS != ret)
  {
    ni_log(NI_LOG_ERROR, "Failed to close Decode Session (status = %d)\n", ret);
  }
#ifdef _WIN32
  ni_logan_device_close(p_session_ctx->device_handle);
#elif __linux__
  ni_logan_device_close(p_session_ctx->device_handle);
  ni_logan_device_close(p_session_ctx->blk_io_handle);
#endif
  p_session_ctx->device_handle = NI_INVALID_DEVICE_HANDLE;
  p_session_ctx->blk_io_handle = NI_INVALID_DEVICE_HANDLE;

  return 0;
}

void ni_logan_dec_close(ni_logan_dec_context_t *ni_dec_ctx)
{
  ni_logan_session_context_t *p_session_ctx = ni_dec_ctx->p_session_ctx;
  ni_logan_decoder_params_t *p_param = ni_dec_ctx->p_param;
  ni_logan_session_data_io_t *p_ni_packet = NULL;
  ni_log(NI_LOG_DEBUG, "ni_logan_dec_close enter\n");

  /* this call shall release resource based on ni_logan_dec_context_t*/
  ni_logan_xcoder_dec_close(ni_dec_ctx);

  free(ni_dec_ctx->extradata);
  ni_dec_ctx->extradata = NULL;
  ni_dec_ctx->extradata_size = 0;
  ni_dec_ctx->got_first_key_frame = 0;
  ni_dec_ctx->dev_xcoder = NULL; //Set to Null
  ni_logan_packet_buffer_free(&ni_dec_ctx->seq_hdr_pkt);

  if (ni_dec_ctx->input_data_fifo)
  {
    // push back
    do{
      p_ni_packet = ni_logan_fifo_generic_write(ni_dec_ctx->input_data_fifo);
    }while(p_ni_packet);
    // free all
    do{
      p_ni_packet = ni_logan_fifo_generic_read(ni_dec_ctx->input_data_fifo);
      if (p_ni_packet)
      {
        ni_logan_packet_buffer_free(&(p_ni_packet->data.packet));
      }
    }while(p_ni_packet);
  }

  ni_logan_fifo_free(ni_dec_ctx->input_data_fifo);
  ni_log(NI_LOG_DEBUG, "%s(): decoder frame fifo freed.\n", __FUNCTION__);
  // no releaseResource for decoder for now
  ni_logan_device_session_context_clear(p_session_ctx);
  if (p_session_ctx)
  {
    free(p_session_ctx);
  }
  if (p_param)
  {
    free(p_param);
  }
}

int ni_logan_setup_decoder(ni_logan_dec_context_t *ni_dec_ctx)
{
  ni_logan_retcode_t retval = NI_LOGAN_RETCODE_SUCCESS;
  ni_logan_decoder_params_t *p_param = NULL;
  ni_logan_session_context_t *p_session_ctx = NULL;
  ni_log_set_level(ff_to_ni_log_level(ni_dec_ctx->ff_log_level));
  ni_log(NI_LOG_DEBUG, "ni_logan_setup_decoder enter\n");

  if (!ni_dec_ctx)
  {
    ni_log(NI_LOG_ERROR, "ni_dec_ctx is null\n");
    return NI_LOGAN_RETCODE_INVALID_PARAM;
  }
  // Xcoder Session Context
  p_session_ctx = malloc(sizeof(ni_logan_session_context_t));
  if (!p_session_ctx)
  {
    ni_log(NI_LOG_ERROR, "ERROR %d:%s() line %d, alloc session context failed\n",
           NI_ERRNO, __FUNCTION__, __LINE__);
    retval = NI_LOGAN_RETCODE_ERROR_MEM_ALOC;
    LRETURN;
  }
  else
  {
    ni_dec_ctx->p_session_ctx = p_session_ctx;
    ni_logan_device_session_context_init(p_session_ctx);
  }

  if (0 == strcmp(ni_dec_ctx->dev_xcoder, LIST_DEVICES_STR))
  {
    ni_log(NI_LOG_DEBUG, "XCoder: printing out all xcoder devices and "
           "their load, and exit ...\n");
    ni_logan_rsrc_print_all_devices_capability();
    return NI_LOGAN_RETCODE_FAILURE;
  }
  else if (ni_dec_ctx->width > NI_LOGAN_MAX_RESOLUTION_WIDTH ||
           ni_dec_ctx->height > NI_LOGAN_MAX_RESOLUTION_HEIGHT ||
           ni_dec_ctx->width * ni_dec_ctx->height > NI_LOGAN_MAX_RESOLUTION_AREA)
  {
    ni_log(NI_LOG_ERROR, "Error XCoder resolution %dx%d not supported\n",
           ni_dec_ctx->width, ni_dec_ctx->height);
    ni_log(NI_LOG_ERROR, "Max Supported Width: %d Height %d Area %d\n",
           NI_LOGAN_MAX_RESOLUTION_WIDTH, NI_LOGAN_MAX_RESOLUTION_HEIGHT, NI_LOGAN_MAX_RESOLUTION_AREA);
    retval = NI_LOGAN_RETCODE_FAILURE;
    LRETURN;
  }
  ni_log(NI_LOG_DEBUG, "XCoder setup device decoder\n");
  if(ni_dec_ctx->dev_dec_idx < 0)
  {
    // the user did not specify which device to use.
    // we need to auto select the device,
    // we need to know which selection method to use.
    // if the slection method is not in place, we use best model by default
    //
    if (strlen(ni_dec_ctx->dev_xcoder) <= 0)
    {
      strcpy(p_session_ctx->dev_xcoder, BEST_MODEL_LOAD_STR);
    }
    else
    {
      strcpy(p_session_ctx->dev_xcoder, ni_dec_ctx->dev_xcoder);
    }
  }
  ni_dec_ctx->started = 0;
  ni_dec_ctx->got_first_key_frame = 0;

  ni_dec_ctx->offset = 0LL;
  ni_dec_ctx->draining = 0;

  //Xcoder User Configuration
  p_param = malloc(sizeof(ni_logan_decoder_params_t));
  if (!p_param)
  {
    ni_log(NI_LOG_ERROR, "ERROR %d:%s() line %d, alloc session context failed\n",
          NI_ERRNO, __FUNCTION__, __LINE__);
    retval = NI_LOGAN_RETCODE_ERROR_MEM_ALOC;
    LRETURN;
  }
  else
  {
    ni_dec_ctx->p_param = p_param;
    if (ni_logan_decoder_init_default_params(ni_dec_ctx->p_param, ni_dec_ctx->fps_num,
                                            ni_dec_ctx->fps_denom, ni_dec_ctx->bit_rate,
                                            ni_dec_ctx->width, ni_dec_ctx->height) < 0)
    {
      ni_log(NI_LOG_INFO, "Error setting params\n");
      retval = NI_LOGAN_RETCODE_FAILURE;
      LRETURN;
    }
  }

  // Init input fifo
  ni_dec_ctx->input_data_fifo = ni_logan_fifo_initialize(NI_LOGAN_DEC_PKT_BUF_NUM_MAX,
                                                         sizeof(ni_logan_session_data_io_t));
  if (!ni_dec_ctx->input_data_fifo)
  {
    ni_log(NI_LOG_ERROR, "%s(): line %d initialize enc frame buffer pool failed\n",
           __FUNCTION__, __LINE__);
    retval = NI_LOGAN_RETCODE_ERROR_MEM_ALOC;
    LRETURN;
  }

END:

  if (retval != NI_LOGAN_RETCODE_SUCCESS)
  {
    free(p_session_ctx);
    ni_dec_ctx->p_session_ctx = NULL;
    free(p_param);
    ni_dec_ctx->p_param = NULL;
    ni_logan_fifo_free(ni_dec_ctx->input_data_fifo);
  }

  return retval;
}

int ni_logan_dec_init(ni_logan_dec_context_t *ni_dec_ctx)
{
  /* ToDo: call xcode_dec_open to open a decoder instance */
  int ret;
  ni_logan_decoder_params_t *p_param = NULL;
  ni_logan_session_context_t *p_session_ctx = NULL;
  ni_log(NI_LOG_DEBUG, "ni_logan_dec_init enter\n");

  if (!ni_dec_ctx)
  {
    ni_log(NI_LOG_ERROR, "ni_dec_ctx is null\n");
    return NI_LOGAN_RETCODE_INVALID_PARAM;
  }
  p_param = ni_dec_ctx->p_param;
  p_session_ctx = ni_dec_ctx->p_session_ctx;
  if (!p_param || !p_session_ctx)
  {
    ni_log(NI_LOG_ERROR, "%s(): line %d, Invalid Parameter, %p, %p\n",
           __FUNCTION__, __LINE__, p_param, p_session_ctx);
    return NI_LOGAN_RETCODE_INVALID_PARAM;
  }

  p_session_ctx->codec_format = ni_dec_ctx->codec_format;
  p_session_ctx->bit_depth_factor = ni_dec_ctx->bit_depth_factor;
  p_session_ctx->pic_reorder_delay = ni_dec_ctx->pic_reorder_delay;

  // 1. overwrite keep_alive_timeout value here with a custom value if it was provided
  uint32_t xcoder_timeout = p_param->dec_input_params.keep_alive_timeout;
  if (xcoder_timeout != NI_LOGAN_DEFAULT_KEEP_ALIVE_TIMEOUT) 
  {
    p_session_ctx->keep_alive_timeout = xcoder_timeout;
  } 
  else 
  {
    p_session_ctx->keep_alive_timeout = ni_dec_ctx->keep_alive_timeout;
  }
  ni_log(NI_LOG_DEBUG, "Custom NVME Keep Alive Timeout set to %d\n",
         p_session_ctx->keep_alive_timeout);

  // 2.overwrite set_high_priority value here with a custom value if it was provided
  uint32_t xcoder_high_priority = p_param->dec_input_params.set_high_priority;
  if(xcoder_high_priority != 0)
  {
    p_session_ctx->set_high_priority = xcoder_high_priority;
  }
  else
  {
    p_session_ctx->set_high_priority = ni_dec_ctx->set_high_priority;
  }
  ni_log(NI_LOG_DEBUG, "Custom NVMe set_high_priority set to = %d\n",
         p_session_ctx->set_high_priority);
  
  // 3. overwrite UDU passthru value here with a custom value if it was provided
  if (p_param->dec_input_params.enable_user_data_sei_passthru != 0)
  {
    ni_dec_ctx->enable_user_data_sei_passthru = p_param->dec_input_params.enable_user_data_sei_passthru;
  }

  // 4. overwrite check_packet value here with a custom value if it was provided
  if (p_param->dec_input_params.check_packet != 0)
  {
    ni_dec_ctx->enable_check_packet = p_param->dec_input_params.check_packet;
  }
  else
  {
    p_param->dec_input_params.check_packet = ni_dec_ctx->enable_check_packet;
  }

  // 5. overwrite UDU passthru value here with a custom value if it was provided
  if (p_param->dec_input_params.custom_sei_passthru != NI_LOGAN_INVALID_SEI_TYPE)
  {
    ni_dec_ctx->custom_sei = p_param->dec_input_params.custom_sei_passthru;
  }
  else
  {
    p_param->dec_input_params.custom_sei_passthru = ni_dec_ctx->custom_sei;
  }

  p_session_ctx->p_session_config = p_param;

  p_session_ctx->hw_id = ni_dec_ctx->dev_dec_idx;
  p_session_ctx->hw_name = ni_dec_ctx->dev_dec_name;
  if(p_param->dec_input_params.lowdelay != 0)
  {
    ni_dec_ctx->low_delay = p_param->dec_input_params.lowdelay;
  }
  else
  {
    p_param->dec_input_params.lowdelay = ni_dec_ctx->low_delay;
  }
  p_session_ctx->decoder_low_delay = ni_dec_ctx->low_delay;
  strcpy(p_session_ctx->dev_xcoder, ni_dec_ctx->dev_xcoder);

  ret = ni_logan_device_session_open(p_session_ctx, NI_LOGAN_DEVICE_TYPE_DECODER);
  if (ret != 0)
  {
    ni_log(NI_LOG_ERROR, "Failed to open decoder (status = %d), "
           "resource unavailable\n", ret);
    ret = NI_LOGAN_RETCODE_FAILURE;
    ni_logan_dec_close(ni_dec_ctx);
    ni_log(NI_LOG_ERROR, "Failed to open decoder (status = %d), "
           "resource unavailable\n", ret);
  }
  else
  {
    ni_dec_ctx->dev_xcoder_name = p_session_ctx->dev_xcoder_name;
    ni_dec_ctx->blk_xcoder_name = p_session_ctx->blk_xcoder_name;
    ni_dec_ctx->dev_dec_idx = p_session_ctx->hw_id;
    ni_log(NI_LOG_DEBUG, "XCoder %s Index %d (inst: %d) opened successfully\n",
           ni_dec_ctx->dev_xcoder_name, ni_dec_ctx->dev_dec_idx, p_session_ctx->session_id);
  }

  ni_dec_ctx->eos_pkt_received = 0;
  ni_dec_ctx->dec_wait = NI_LOGAN_POLL_INTERVAL;
  ni_dec_ctx->current_pts = 0;
  return ret;
}

int ni_logan_dec_reinit(ni_logan_dec_context_t *ni_dec_ctx)
{
  int ret = 0;
  ni_log(NI_LOG_INFO, "ni_logan_dec_reinit enter\n");
  int64_t bcp_current_pts = ni_dec_ctx->current_pts;
  if ((ret = ni_logan_setup_decoder(ni_dec_ctx)) < 0)
  {
    return ret;
  }
  ni_logan_session_context_t *p_session_ctx = ni_dec_ctx->p_session_ctx;
  if ((ret = ni_logan_dec_init(ni_dec_ctx)) < 0)
  {
    goto done;
  }
  ni_dec_ctx->current_pts = bcp_current_pts;
done:
  if (NI_INVALID_DEVICE_HANDLE == p_session_ctx->blk_io_handle ||
      NI_INVALID_DEVICE_HANDLE == p_session_ctx->device_handle)
  {
    ni_logan_dec_close(ni_dec_ctx);
  }

  return ret;
}
// return 1 if need to prepend saved header to pkt data, 0 otherwise
int ni_logan_xcoder_add_headers(ni_logan_dec_context_t *ni_dec_ctx, uint8_t *pkt_data, int pkt_size,
                                int pkt_flags, uint8_t *extradata, int extradata_size)
{
  int ret = 0;
  const uint8_t *ptr = pkt_data;
  const uint8_t *end = pkt_data + pkt_size;
  uint32_t stc = -1;
  uint8_t nalu_type = 0;

  if (!pkt_data)
  {
    return ret;
  }

  while (ptr < end)
  {
    stc = -1;
    ptr = ni_logan_find_start_code(ptr, end, &stc);
    if (ptr == end)
    {
      break;
    }

    // NAL type constant names have conflicts between AVC and HEVC, so
    // use numbers directly for AVC
    if (NI_LOGAN_CODEC_FORMAT_H264 == ni_dec_ctx->codec_format)
    {
      nalu_type = stc & 0x1F;

      // update extra data on sequence change by resetting got_first_key_frame
      if (ni_dec_ctx->got_first_key_frame && pkt_flags & 0x0001)
      {
        if (ni_dec_ctx->extradata_size != extradata_size ||
            memcmp(ni_dec_ctx->extradata, extradata, ni_dec_ctx->extradata_size))
        {
          ni_dec_ctx->got_first_key_frame = 0;
        }
      }

      // Find the first packet containing key frame
      if (!ni_dec_ctx->got_first_key_frame && pkt_flags & 0x0001)
      {
        free(ni_dec_ctx->extradata);
        ni_dec_ctx->extradata = malloc(extradata_size);
        if (!ni_dec_ctx->extradata)
        {
          ni_log(NI_LOG_ERROR, "%s memory allocation failed !\n", __FUNCTION__);
          ret = 0;
          break;
        }
        ni_log(NI_LOG_TRACE, "%s size %d\n", __FUNCTION__, extradata_size);
        memcpy(ni_dec_ctx->extradata, extradata, extradata_size);
        ni_dec_ctx->extradata_size = extradata_size;
        ni_dec_ctx->got_first_key_frame = 1;
        ret = 1;
        break;
      }

      // when header (SPS/PPS) already exists, no need to prepend it again;
      // we use one of the header info to simplify the checking.
      if (/*NAL_SPS*/7 == nalu_type || /*NAL_PPS*/8 == nalu_type)
      {
        // save the header if not done yet for subsequent comparison
        if (! ni_dec_ctx->extradata_size || ! ni_dec_ctx->extradata)
        {
          ni_dec_ctx->extradata = malloc(extradata_size);
          if (! ni_dec_ctx->extradata)
          {
            ni_log(NI_LOG_ERROR, "%s memory allocation failed!\n",
                   __FUNCTION__);
            ret = 0;
            break;
          }
          ni_log(NI_LOG_TRACE, "%s size %d\n",
                 __FUNCTION__, extradata_size);
          memcpy(ni_dec_ctx->extradata, extradata, extradata_size);
          ni_dec_ctx->extradata_size = extradata_size;
        }
        ni_dec_ctx->got_first_key_frame = 1;
        ret = 0;
        break;
      }
      else if (nalu_type >= 1/*NAL_SLICE*/ && nalu_type <= 5/*NAL_IDR_SLICE*/)
      {
        // VCL types results in no header inserted
        ret = 0;
        break;
      }
    }
    else if (NI_LOGAN_CODEC_FORMAT_H265 == ni_dec_ctx->codec_format)
    {
      nalu_type = (stc >> 1) & 0x3F;

      // IRAP picture types include: BLA, CRA, IDR and IRAP reserve types,
      // 16-23, and insert header in front of IRAP at start or if header changes
      if (ni_dec_ctx->got_first_key_frame && pkt_flags & 0x0001)
      {
        if (ni_dec_ctx->extradata_size != extradata_size ||
            memcmp(ni_dec_ctx->extradata, extradata, ni_dec_ctx->extradata_size))
        {
          ni_dec_ctx->got_first_key_frame = 0;
        }
      }

      if (! ni_dec_ctx->got_first_key_frame && pkt_flags & 0x0001)
      {
        free(ni_dec_ctx->extradata);
        ni_dec_ctx->extradata = malloc(extradata_size);
        if (! ni_dec_ctx->extradata)
        {
          ni_log(NI_LOG_ERROR, "%s memory allocation failed !\n",
                 __FUNCTION__);
          ret = 0;
          break;
        }
        ni_log(NI_LOG_TRACE, "%s size %d\n",
               __FUNCTION__, extradata_size);
        memcpy(ni_dec_ctx->extradata, extradata, extradata_size);
        ni_dec_ctx->extradata_size = extradata_size;
        ni_dec_ctx->got_first_key_frame = 1;
        ret = 1;
        break;
      }

      // when header (VPS/SPS/PPS) already exists, no need to prepend it again;
      // we use one of the header info to simplify the checking.
      if (/*NI_NAL_VPS*/32 == nalu_type || /*NI_NAL_SPS*/33 == nalu_type || /*NI_NAL_PPS*/34 == nalu_type)
      {
        // save the header if not done yet for subsequent comparison
        if (! ni_dec_ctx->extradata_size || ! ni_dec_ctx->extradata)
        {
          ni_dec_ctx->extradata = malloc(extradata_size);
          if (! ni_dec_ctx->extradata)
          {
            ni_log(NI_LOG_ERROR, "%s memory allocation failed!\n",
                   __FUNCTION__);
            ret = 0;
            break;
          }
          ni_log(NI_LOG_TRACE, "%s size %d\n",
                 __FUNCTION__, extradata_size);
          memcpy(ni_dec_ctx->extradata, extradata, extradata_size);
          ni_dec_ctx->extradata_size = extradata_size;
        }
        ni_dec_ctx->got_first_key_frame = 1;
        ret = 0;
        break;
      }
      else if (nalu_type >= 0/*NI_NAL_TRAIL_N*/ && nalu_type <= 31)
      {
        // VCL types results in no header inserted
        ret = 0;
        break;
      }
    }
    else
    {
      ni_log(NI_LOG_ERROR, "%s wrong codec %d !\n",
             __FUNCTION__, ni_dec_ctx->codec_format);
      break;
    }
  }

  return ret;
}


// check if the packet is SEI only, 1 yes, 0 otherwise
// check if getting the header of streams in decoder low delay mode, and update its value
// check the new sequence headers and cache them.
int ni_logan_packet_parse(ni_logan_dec_context_t *ni_dec_ctx, uint8_t *pkt_data,
                          int pkt_size, ni_logan_packet_t *p_packet)
{
  ni_logan_session_context_t *p_session_ctx = ni_dec_ctx->p_session_ctx;
  ni_logan_decoder_params_t *p_param = ni_dec_ctx->p_param;
  int pkt_sei_alone = 0;
  int got_slice = 0;
  int low_delay = ni_dec_ctx->low_delay;
  int pkt_nal_bitmap = 0;
  int chk_pkt = ni_dec_ctx->enable_check_packet;
  const uint8_t *ptr = pkt_data;
  const uint8_t *end = pkt_data + pkt_size;
  uint8_t nalu_type = 0;
  int nalu_count = 0;
  int stc;

  if (!pkt_data)
  {
    return pkt_sei_alone;
  }

  if (ni_dec_ctx->pkt_nal_bitmap & NI_LOGAN_GENERATE_ALL_NAL_HEADER_BIT)
  {
    ni_log(NI_LOG_TRACE, "%s(): already find the header of streams.\n",
           __FUNCTION__);
    low_delay = 0;
  }
  pkt_sei_alone = 1;
  while ((pkt_sei_alone || low_delay || chk_pkt) && ptr < end)
  {
    stc = -1;
    ptr = ni_logan_find_start_code(ptr, end, &stc);
    if (ptr == end)
    {
      if (0 == nalu_count)
      {
        pkt_sei_alone = 0;
        ni_log(NI_LOG_TRACE, "%s(): no NAL found in pkt\n", __FUNCTION__);
      }
      break;
    }
    nalu_count++;

    if (NI_LOGAN_CODEC_FORMAT_H264 == ni_dec_ctx->codec_format)
    {
      nalu_type = stc & 0x1F;

      //check whether the packet is sei alone
      pkt_sei_alone = (pkt_sei_alone && 6/*NAL_SEI*/ == nalu_type);

      //check whether the packet contains SEI NAL after VCL NAL units
      if (got_slice && (6/*NAL_SEI*/ == nalu_type))
      {
        chk_pkt = 0;
        // 5 bytes = 3 bytes start code + 1 byte nal type(0x06) + 1 byte sei type
        p_packet->len_of_sei_after_vcl = (end - ptr) + 5;
        ni_log(NI_LOG_TRACE, "%s(): found SEI after VCL NAL, len=%d\n",
               __FUNCTION__, p_packet->len_of_sei_after_vcl);
      }
      else if ((nalu_type >= 1/*NAL_SLICE*/) && (nalu_type <= 5/*NAL_IDR_SLICE*/)) //VCL
      {
        got_slice = 1;
      }

      // FFmpeg stores a complete frame in each AVPacket. A complete frame
      // includes headers, SEIs, video slices, and other NALs such as access
      // unit delimiters, etc. The decoder expects a complete frame for each
      // packet writing to the decoder. Otherwise it can cause all sorts of
      // problems. However in some cases the first NALU could be unit delimiter
      // so we need to force to check NALU split to collect all the headers in
      // case of decoder VPU recovery.
      switch (nalu_type)
      {
        case 7/*NAL_SPS*/:
          chk_pkt = 1;
          pkt_nal_bitmap |= NI_LOGAN_NAL_SPS_BIT;
          break;
        case 8/*NAL_PPS*/:
          chk_pkt = 1;
          pkt_nal_bitmap |= NI_LOGAN_NAL_PPS_BIT;
          break;
        case 6/*NAL_SEI*/:
        case 9/*NAL_AUD*/:
          chk_pkt = 1;
          break;
        default:
          chk_pkt = ni_dec_ctx->enable_check_packet | p_param->dec_input_params.check_packet;
          break;
      }

      // Set decoder low delay mode one-time.
      if (pkt_nal_bitmap & (NI_LOGAN_NAL_SPS_BIT | NI_LOGAN_NAL_PPS_BIT))
      {
        ni_log(NI_LOG_TRACE, "%s(): Detect SPS, PPS and IDR, enable "
               "decoder low delay mode.\n", __FUNCTION__);
        pkt_nal_bitmap |= NI_LOGAN_GENERATE_ALL_NAL_HEADER_BIT;
        if (low_delay)
        {
          p_session_ctx->decoder_low_delay = low_delay;
          low_delay = 0;
        }

        // Update cached packet including SPS+PPS+IDR slice. A complete frame
        // needs to be cached for stream with intraPeriod = 0
        if(got_slice)
        {
          p_packet->frame_type = LOGAN_PIC_TYPE_IDR;
        }
      }
    }
    else if (NI_LOGAN_CODEC_FORMAT_H265 == ni_dec_ctx->codec_format)
    {
      nalu_type = (stc >> 1) & 0x3F;

      //check whether the packet is sei alone
      pkt_sei_alone = (pkt_sei_alone && (39 == nalu_type || 40 == nalu_type));

      //check whether the packet contains SEI NAL after VCL NAL units
      if (got_slice && (/*NI_NAL_SEI_PREFIX*/39 == nalu_type || /*NI_NAL_SEI_SUFFIX*/40 == nalu_type))
      {
        chk_pkt = 0;
        // 5 bytes = 3 bytes start code + 2 bytes nal type(0x4e 0x01)
        p_packet->len_of_sei_after_vcl = (end - ptr) + 5;
        ni_log(NI_LOG_TRACE, "%s(): found SEI after VCL NAL, len=%d\n",
               __FUNCTION__, p_packet->len_of_sei_after_vcl);
      }
      else if ((nalu_type >= 0/*NI_NAL_TRAIL_N*/) && (nalu_type <= 31)) //VCL
      {
        got_slice = 1;
      }

      // FFmpeg stores a complete frame in each AVPacket. A complete frame
      // includes headers, SEIs, video slices, and other NALs such as access
      // unit delimiters, etc. The decoder expects a complete frame for each
      // packet writing to the decoder. Otherwise it can cause all sorts of
      // problems. However in some cases the first NALU could be unit delimiter
      // so we need to force to check NALU split to collect all the headers in
      // case of decoder VPU recovery.
      switch (nalu_type)
      {
        case 32/*NI_NAL_VPS*/:
          chk_pkt = 1;
          pkt_nal_bitmap |= NI_LOGAN_NAL_VPS_BIT;
          break;
        case 33/*NI_NAL_SPS*/:
          chk_pkt = 1;
          pkt_nal_bitmap |= NI_LOGAN_NAL_SPS_BIT;
          break;
        case 34/*NI_NAL_PPS*/:
          chk_pkt = 1;
          pkt_nal_bitmap |= NI_LOGAN_NAL_PPS_BIT;
          break;
        case 35/*NI_NAL_AUD*/:
        case 36/*NI_NAL_EOS_NUT*/:
        case 37/*NI_NAL_EOB_NUT*/:
        case 38/*NI_NAL_FD_NUT*/:
        case 39/*NI_NAL_SEI_PREFIX*/:
        case 40/*NI_NAL_SEI_SUFFIX*/:
          chk_pkt = 1;
          break;
        default:
          chk_pkt = ni_dec_ctx->enable_check_packet | p_param->dec_input_params.check_packet;
          break;
      }

      // Set decoder low delay mode one-time.
      if (pkt_nal_bitmap & (NI_LOGAN_NAL_VPS_BIT | NI_LOGAN_NAL_SPS_BIT | NI_LOGAN_NAL_PPS_BIT))
      {
        ni_log(NI_LOG_TRACE, "%s(): Detect VPS, SPS, PPS and IDR, "
               "enable decoder low delay mode.\n", __FUNCTION__);
        pkt_nal_bitmap |= NI_LOGAN_GENERATE_ALL_NAL_HEADER_BIT;
        if (low_delay)
        {
          p_session_ctx->decoder_low_delay = low_delay;
          low_delay = 0;
        }

        // Update cached packet including VPS+SPS+PPS+IDR slice. A complete frame
        // needs to be cached for stream with intraPeriod = 0
        if(got_slice)
        {
          p_packet->frame_type = LOGAN_PIC_TYPE_IDR;
        }
      }
    }
    else
    {
      ni_log(NI_LOG_ERROR, "%s() wrong codec %d!\n",
             __FUNCTION__, ni_dec_ctx->codec_format);
      pkt_sei_alone = 0;
      break;
    }
  }

  ni_dec_ctx->pkt_nal_bitmap |= pkt_nal_bitmap;
  return pkt_sei_alone;
}

int ni_logan_packet_buffer_clone(ni_logan_packet_t *dst_pkt, ni_logan_packet_t *src_pkt)
{
  ni_log(NI_LOG_ERROR,"ni_logan_packet_buffer_clone enter\n");

  dst_pkt->pts = src_pkt->pts;
  dst_pkt->dts = src_pkt->dts;
  dst_pkt->flags = src_pkt->flags;
  dst_pkt->video_width = src_pkt->video_width;
  dst_pkt->video_height = src_pkt->video_height;
  dst_pkt->data_len = src_pkt->data_len;
  dst_pkt->p_data = src_pkt->p_data;
  src_pkt->p_data = NULL;
  return 0;
}

int ni_logan_dec_send(ni_logan_dec_context_t *ni_dec_ctx, uint8_t *pkt_data,
                      int pkt_size, int pkt_flags, long long pkt_pts,
                      long long pkt_dts, int avctx_width, int avctx_height,
                      uint8_t *extradata, int extradata_size)
{
  /* call xcoder_dec_send to send compressed video packet to the decoder
     instance */
  ni_logan_session_context_t *p_session_ctx = ni_dec_ctx->p_session_ctx;
  ni_logan_decoder_params_t *p_param = ni_dec_ctx->p_param;
  ni_logan_session_data_io_t *p_ni_packet = NULL;
  int need_draining = 0;
  size_t size;
  ni_logan_packet_t *xpkt = NULL;
  int ret;
  int sent;
  int send_size = 0;
  int new_packet = 0;
  int extra_prev_size = 0;

  size = pkt_size;

  if (ni_dec_ctx->flushing)
  {
    ni_log(NI_LOG_ERROR, "Decoder is flushing and cannot accept new "
           "buffer until all output buffers have been released\n");
    return NI_LOGAN_RETCODE_FAILURE;
  }

  if (pkt_size == 0)
  {
    need_draining = 1;
    // Once VPU recovery, the s->draining may be lost during session reset. And
    // so is the EOS indicator which might be lost in decoder session read. Here
    // we try to recover EOS indicator so as to return EOF in this calling.
    ni_dec_ctx->eos |= (ni_dec_ctx->draining && 
                        (p_session_ctx->session_run_state == LOGAN_SESSION_RUN_STATE_RESETTING));
    ni_log(NI_LOG_DEBUG, "Decoder is need_draining= %d , draining=%d,session_run_state=%d,eos=%d\n",
           need_draining,ni_dec_ctx->draining,p_session_ctx->session_run_state,ni_dec_ctx->eos);
  }

  if (ni_dec_ctx->draining && ni_dec_ctx->eos)
  {
    ni_log(NI_LOG_DEBUG, "Decoder is draining, eos\n");
    return NI_LOGAN_RETCODE_FAILURE;
  }
  if (!ni_dec_ctx->eos_pkt_received)
  {
    if (ni_logan_fifo_is_full(ni_dec_ctx->input_data_fifo))
    {
      ni_log(NI_LOG_ERROR,"fifo full,need lost 1 frame....\n");
      p_ni_packet = ni_logan_fifo_generic_read(ni_dec_ctx->input_data_fifo);
      ni_logan_packet_buffer_free(&(p_ni_packet->data.packet));
    }
    p_ni_packet = ni_logan_fifo_generic_write(ni_dec_ctx->input_data_fifo);
    xpkt = &p_ni_packet->data.packet;
    {
      memset(xpkt, 0, sizeof(ni_logan_packet_t));
      xpkt->pts = pkt_pts;
      xpkt->dts = pkt_dts;
      xpkt->flags = pkt_flags;
      xpkt->video_width = avctx_width;
      xpkt->video_height = avctx_height;
      xpkt->p_data = NULL;
      xpkt->data_len = pkt_size;
      xpkt->p_all_custom_sei = NULL;
      xpkt->len_of_sei_after_vcl = 0;


      if (extradata_size > 0 && extradata_size != ni_dec_ctx->extradata_size &&
          ni_logan_xcoder_add_headers(ni_dec_ctx, pkt_data, pkt_size, pkt_flags, extradata, extradata_size))
      {
        if (extradata_size > p_session_ctx->max_nvme_io_size * 2)
        {
          ni_log(NI_LOG_ERROR, "%s extradata_size %d exceeding max size "
                "supported: %d\n", __FUNCTION__, extradata_size,
                p_session_ctx->max_nvme_io_size * 2);
        }
        else
        {
          ni_log(NI_LOG_DEBUG, "%s extradata_size %d copied to pkt start\n",
                __FUNCTION__, extradata_size);
          p_session_ctx->prev_size = extradata_size;
          memcpy(p_session_ctx->p_leftover, extradata, extradata_size);
        }
      }

      if (ni_dec_ctx->low_delay  && ni_dec_ctx->got_first_key_frame &&
          !(ni_dec_ctx->pkt_nal_bitmap & NI_LOGAN_GENERATE_ALL_NAL_HEADER_BIT))
      {
        p_session_ctx->decoder_low_delay = p_param->dec_input_params.lowdelay;
        ni_dec_ctx->pkt_nal_bitmap |= NI_LOGAN_GENERATE_ALL_NAL_HEADER_BIT;
        ni_log(NI_LOG_TRACE, "%s got first IDR in decoder low delay mode, "
              "delay time %dms, pkt_nal_bitmap %d\n", __FUNCTION__,
              p_session_ctx->decoder_low_delay, ni_dec_ctx->pkt_nal_bitmap);
      }

      // check if the packet is SEI only, save it to be sent with the next data frame.
      // check if getting the header of streams in decoder low delay mode, and update its value.
      // check the new sequence headers and cache them.
      ret = ni_logan_packet_parse(ni_dec_ctx, pkt_data, pkt_size, xpkt);
      if (ret)
      {
        // skip the packet if it's corrupted and/or exceeding lone SEI buf size
        if (pkt_size + p_session_ctx->lone_sei_size <= NI_LOGAN_MAX_SEI_DATA)
        {
          memcpy(p_session_ctx->buf_lone_sei + p_session_ctx->lone_sei_size,
                pkt_data, pkt_size);
          p_session_ctx->lone_sei_size += pkt_size;

          ni_log(NI_LOG_TRACE, "%s pkt lone SEI saved and return %d\n",
                __FUNCTION__, pkt_size);
        }
        else
        {
          ni_log(NI_LOG_DEBUG, "lone SEI size %d > buf size %ld "
                "corrupted? skipped ..\n", pkt_size, NI_LOGAN_MAX_SEI_DATA);
        }

        xpkt->data_len = 0;
        return pkt_size;
      }

      if (ni_dec_ctx->custom_sei != NI_LOGAN_INVALID_SEI_TYPE)
      {
        ret = ni_logan_detect_custom_sei(pkt_data, pkt_size, p_session_ctx, p_param, xpkt);
        if (ret != 0)
        {
          goto fail;
        }
      }

      // embed lone SEI saved previously (if any) to send to decoder
      if (p_session_ctx->lone_sei_size)
      {
        ni_log(NI_LOG_TRACE, "%s copy over lone SEI data size: %d\n",
              __FUNCTION__, p_session_ctx->lone_sei_size);
        memcpy((uint8_t *)p_session_ctx->p_leftover + p_session_ctx->prev_size,
              p_session_ctx->buf_lone_sei, p_session_ctx->lone_sei_size);
        p_session_ctx->prev_size += p_session_ctx->lone_sei_size;
        p_session_ctx->lone_sei_size = 0;
      }

      if ((pkt_size + p_session_ctx->prev_size) > 0)
      {
        ni_logan_packet_buffer_alloc(xpkt, (pkt_size + p_session_ctx->prev_size - xpkt->len_of_sei_after_vcl));
        if (!xpkt->p_data)
        {
          ret = NI_LOGAN_RETCODE_ERROR_MEM_ALOC;
          goto fail;
        }
      }
      new_packet = 1;
    }

    ni_log(NI_LOG_DEBUG, "%s: pkt size=%d\n", __FUNCTION__, pkt_size);

    if (ni_dec_ctx->started == 0)
    {
      xpkt->start_of_stream = 1;
      ni_dec_ctx->started = 1;
    }

    ni_log(NI_LOG_TRACE, "ni_logan_packet_copy before: size=%d s->prev_size=%d"
           " send_size=%d len_of_sei_after_slice=%d (end of stream)\n",
           pkt_size, p_session_ctx->prev_size, send_size, xpkt->len_of_sei_after_vcl);
    if (new_packet)
    {
      extra_prev_size = p_session_ctx->prev_size;
      send_size = ni_logan_packet_copy(xpkt->p_data, pkt_data, (pkt_size - xpkt->len_of_sei_after_vcl),
                                  p_session_ctx->p_leftover, &p_session_ctx->prev_size);
      // increment offset of data sent to decoder and save it
      xpkt->pos = ni_dec_ctx->offset;
      ni_dec_ctx->offset += pkt_size + extra_prev_size;
    }
    ni_log(NI_LOG_TRACE, "ni_logan_packet_copy after: size=%d, s->prev_size=%d"
           " send_size=%d xpkt->data_len=%d, len_of_sei_after_slice=%d (end of "
           "stream)\n", pkt_size, p_session_ctx->prev_size, send_size,
           xpkt->data_len, xpkt->len_of_sei_after_vcl);

    if (send_size < 0)
    {
      ni_log(NI_LOG_ERROR, "Failed to copy pkt (status = %d)\n",
              send_size);
      ret = NI_LOGAN_RETCODE_FAILURE;
      goto fail;
    }
    xpkt->data_len += extra_prev_size;
    if (need_draining && !ni_dec_ctx->draining)
    {
      ni_log(NI_LOG_DEBUG, "Sending End Of Stream signal\n");
      xpkt->end_of_stream = 1;
      xpkt->data_len = 0;
      ni_dec_ctx->eos_pkt_received = 1;
    }
  }

  if (ni_dec_ctx->flush_buffer == NI_LOGAN_DECODE_FLUSH_STATUS_START &&
     ni_dec_ctx->dec_fifo_num == 0)
  {
    ni_logan_decoder_session_flush_buffers(p_session_ctx);
    ni_dec_ctx->flush_buffer = NI_LOGAN_DECODE_FLUSH_STATUS_FLUSHING;
    return 0;
  }
  else if (ni_dec_ctx->flush_buffer == NI_LOGAN_DECODE_FLUSH_STATUS_FLUSHING)
  {
    if (p_session_ctx->ready_to_flush == NI_LOGAN_DECODE_FLUSH_BUFFER_OFF)
      ni_dec_ctx->flush_buffer = NI_LOGAN_DECODE_FLUSH_STATUS_OFF;
    else
      return 0;
  }

  if ((ni_dec_ctx->eos_pkt_received && !ni_dec_ctx->draining) ||
     ni_dec_ctx->flush_buffer == NI_LOGAN_DECODE_FLUSH_STATUS_START)
  { 
    sent = 1;
    while (!ni_logan_fifo_is_empty(ni_dec_ctx->input_data_fifo))
    {
      p_ni_packet = ni_logan_fifo_generic_read(ni_dec_ctx->input_data_fifo);
      ni_log(NI_LOG_TRACE, "%s pts=%" PRIi64 ", dts=%" PRIi64"\n", __FUNCTION__,
             p_ni_packet->data.packet.pts, p_ni_packet->data.packet.dts);

      sent = ni_logan_device_session_write(p_session_ctx, &(p_ni_packet->data.packet), NI_LOGAN_DEVICE_TYPE_DECODER);

      if (sent == 0 && !ni_logan_fifo_is_empty(ni_dec_ctx->input_data_fifo))
      {
        ni_log(NI_LOG_DEBUG, "dec_send EAGAIN this better not happen "
                                      "sleep %d us !\n",
              ni_dec_ctx->dec_wait);
        usleep(ni_dec_ctx->dec_wait);
        ni_dec_ctx->dec_wait += NI_LOGAN_POLL_INTERVAL;
        ni_logan_fifo_return_read(ni_dec_ctx->input_data_fifo);
        break;
      }
      else if (sent < 0)
      {
        ni_log(NI_LOG_ERROR, "Failed to send eos signal (status = %d)\n",sent);
        if (NI_LOGAN_RETCODE_ERROR_VPU_RECOVERY == sent)
        {
          ni_log(NI_LOG_INFO, "%s VPU recovery, need to reset\n",
                __FUNCTION__);
          p_session_ctx->session_run_state = LOGAN_SESSION_RUN_STATE_RESETTING;

          if (ni_logan_decode_reset(ni_dec_ctx) < 0)
          {
            return NI_LOGAN_RETCODE_FAILURE;
          }
          if (ni_dec_ctx->seq_hdr_pkt.data_len > 0)
          {
            if (ni_logan_device_session_write(p_session_ctx, &ni_dec_ctx->seq_hdr_pkt, NI_LOGAN_DEVICE_TYPE_DECODER) < 0)
            {
              p_session_ctx->session_run_state = LOGAN_SESSION_RUN_STATE_NORMAL;
              return NI_LOGAN_RETCODE_FAILURE;
            }
          }
          return 0;
        }
        ret = NI_LOGAN_RETCODE_FAILURE;
        goto fail;
      }
      else
      {
        if (&(p_ni_packet->data.packet).frame_type == LOGAN_PIC_TYPE_IDR)
        {
          if (ni_dec_ctx->seq_hdr_pkt.data_len > 0)
          {
            ni_logan_packet_buffer_free(&(p_ni_packet->data.packet));
          }
          ni_logan_packet_buffer_clone(&ni_dec_ctx->seq_hdr_pkt, &(p_ni_packet->data.packet));
        }
        ni_logan_packet_buffer_free(&(p_ni_packet->data.packet));
        if (&(p_ni_packet->data.packet).data_len == 0)
        {
          free(&(p_ni_packet->data.packet).p_all_custom_sei);
          (p_ni_packet->data.packet).p_all_custom_sei = NULL;
        }
        // wait time is incremental, and reset after a successful send
        ni_dec_ctx->dec_wait = NI_LOGAN_POLL_INTERVAL;
        if (ni_dec_ctx->flush_buffer == NI_LOGAN_DECODE_FLUSH_STATUS_START)
        {
          ni_dec_ctx->dec_fifo_num --;
          if (ni_dec_ctx->dec_fifo_num == 0)
          {
            ni_logan_decoder_session_flush_buffers(p_session_ctx);
            ni_dec_ctx->flush_buffer = NI_LOGAN_DECODE_FLUSH_STATUS_FLUSHING;
            return 0;
          }
        }
      }
    }
    if (ni_logan_fifo_is_empty(ni_dec_ctx->input_data_fifo))
    {
      ni_logan_device_session_flush(p_session_ctx, NI_LOGAN_DEVICE_TYPE_DECODER);
      ni_dec_ctx->draining = 1;
      p_session_ctx->session_run_state = LOGAN_SESSION_RUN_STATE_NORMAL;
      ni_dec_ctx->eos_pkt_received = 0;
    }
  }
  else
  {
    sent = 0;

    p_ni_packet = ni_logan_fifo_generic_read(ni_dec_ctx->input_data_fifo);
    ni_log(NI_LOG_DEBUG, "%s pts=%" PRIi64 ", dts=%" PRIi64 " \n", __FUNCTION__, p_ni_packet->data.packet.pts, p_ni_packet->data.packet.dts);

    sent = ni_logan_device_session_write(p_session_ctx, &(p_ni_packet->data.packet), NI_LOGAN_DEVICE_TYPE_DECODER);
    if (sent < 0)
    {
      ni_log(NI_LOG_ERROR, "Failed to send compressed pkt (status=%d)\n", sent);
      if (NI_LOGAN_RETCODE_ERROR_VPU_RECOVERY == sent)
      {
        ni_log(NI_LOG_INFO, "%s VPU recovery, need to reset\n",
              __FUNCTION__);
        p_session_ctx->session_run_state = LOGAN_SESSION_RUN_STATE_RESETTING;

        if (ni_logan_decode_reset(ni_dec_ctx) < 0)
        {
          return NI_LOGAN_RETCODE_FAILURE;
        }
        if (ni_dec_ctx->seq_hdr_pkt.data_len > 0)
        {
          if (ni_logan_device_session_write(p_session_ctx, &ni_dec_ctx->seq_hdr_pkt, NI_LOGAN_DEVICE_TYPE_DECODER) < 0)
          {
            p_session_ctx->session_run_state = LOGAN_SESSION_RUN_STATE_NORMAL;
            return NI_LOGAN_RETCODE_FAILURE;
          }
        }
        return 0;
      }
      ret = NI_LOGAN_RETCODE_FAILURE;
      goto fail;
    }
    else if (sent == 0)
    {
      ni_logan_fifo_return_read(ni_dec_ctx->input_data_fifo);
      sent = size;
    }
    else
    {
      if (&(p_ni_packet->data.packet).frame_type == LOGAN_PIC_TYPE_IDR)
      {
        if (ni_dec_ctx->seq_hdr_pkt.data_len > 0)
        {
          ni_logan_packet_buffer_free(&(p_ni_packet->data.packet));
        }
        ni_logan_packet_buffer_clone(&ni_dec_ctx->seq_hdr_pkt, &(p_ni_packet->data.packet));
      }
      ni_logan_packet_buffer_free(&(p_ni_packet->data.packet));
      if (&(p_ni_packet->data.packet).data_len == 0)
      {
        free(&(p_ni_packet->data.packet).p_all_custom_sei);
        (p_ni_packet->data.packet).p_all_custom_sei = NULL;
      }
    }
  }

  return sent;

fail:
  ni_logan_packet_buffer_free(&(p_ni_packet->data.packet));
  free(&(p_ni_packet->data.packet).p_all_custom_sei);
  (p_ni_packet->data.packet).p_all_custom_sei = NULL;
  ni_dec_ctx->draining = 1;
  ni_dec_ctx->eos = 1;
  
  return ret;
}

int ni_logan_dec_flush(ni_logan_dec_context_t *ni_dec_ctx)
{
  ni_dec_ctx->draining = 0;
  ni_dec_ctx->flushing = 0;
  ni_dec_ctx->eos = 0;
  /* Future: for now, always return 1 to indicate the codec has been flushed
     and it leaves the flushing state and can process again ! will consider
     case of user retaining frames in HW "surface" usage */
  return 1;
}

int ni_logan_dec_is_flushing(ni_logan_dec_context_t *ni_dec_ctx)
{
  int ret = 1;
  if (ni_dec_ctx->flushing)
  {
    if (!ni_logan_dec_flush(ni_dec_ctx))
    {
      ni_log(NI_LOG_DEBUG, "IS flushing and NOT dec_flush, AGAIN!!!\n");
      ret = 0;
    }
  }
  return ret;
}

int ni_logan_decoder_frame_alloc(ni_logan_dec_context_t *ni_dec_ctx,
                                 ni_logan_session_data_io_t *p_session_data,
                                 int avctx_width,
                                 int avctx_height)
{
  ni_logan_session_context_t *p_session_ctx = ni_dec_ctx->p_session_ctx;
  int ret, width, height, alloc_mem;

  // If active video resolution has been obtained we just use it as it's the
  // exact size of frame to be returned, otherwise we use what we are told by
  // upper stream as the initial setting and it will be adjusted.
  width = p_session_ctx->active_video_width > 0 ? p_session_ctx->active_video_width : avctx_width;
  height = p_session_ctx->active_video_height > 0 ? p_session_ctx->active_video_height : avctx_height;

  // allocate memory only after resolution is known (buffer pool set up)
  alloc_mem = (p_session_ctx->active_video_width > 0) &&
              (p_session_ctx->active_video_height > 0 ? 1 : 0);

  ret = ni_logan_decoder_frame_buffer_alloc(p_session_ctx->dec_fme_buf_pool,
                                            &p_session_data->data.frame,
                                            alloc_mem,
                                            width,
                                            height,
                                            ni_dec_ctx->codec_format == NI_LOGAN_CODEC_FORMAT_H264,
                                            p_session_ctx->bit_depth_factor);

  return ret;
}

int ni_logan_dec_receive(ni_logan_dec_context_t *ni_dec_ctx,
                            ni_logan_session_data_io_t *p_session_data,
                            int avctx_width,
                            int avctx_height)
{
  ni_logan_session_context_t *p_session_ctx = ni_dec_ctx->p_session_ctx;
  int ret;
  ret = ni_logan_decoder_frame_alloc(ni_dec_ctx, p_session_data, avctx_width, avctx_height);
  if (NI_LOGAN_RETCODE_SUCCESS != ret)
  {
    return ret;
  }

  ret = ni_logan_device_session_read(p_session_ctx, p_session_data, NI_LOGAN_DEVICE_TYPE_DECODER);
  if (ret == 0)
  {
    ni_dec_ctx->eos = p_session_data->data.frame.end_of_stream;
    ni_logan_decoder_frame_free(p_session_data);
    return ret;
  }
  else if (ret > 0) //This is where to handle the frame change and all reinit i guess in revB
  {
    if (p_session_ctx->session_run_state == LOGAN_SESSION_RUN_STATE_RESETTING)
    {
      // On decoder VPU recovery the first received frame corresponding to the
      // cached seq_hdr_pkt should be dropped since the data is outdated.
      p_session_ctx->session_run_state = LOGAN_SESSION_RUN_STATE_NORMAL;
      ni_logan_decoder_frame_free(p_session_data);
      return 0;
    }
  }
  else
  {
    ni_log(NI_LOG_ERROR, "Failed to get output buffer (status=%d)\n", ret);
    if (NI_LOGAN_RETCODE_ERROR_VPU_RECOVERY == ret)
    {
      ni_log(NI_LOG_INFO, "%s VPU recovery, need to reset\n",
             __FUNCTION__);
      p_session_ctx->session_run_state = LOGAN_SESSION_RUN_STATE_RESETTING;
      ni_logan_decoder_frame_free(p_session_data);

      if (ni_logan_decode_reset(ni_dec_ctx) < 0)
      {
        return NI_LOGAN_RETCODE_FAILURE;
      }
      if (ni_dec_ctx->seq_hdr_pkt.data_len > 0)
      {
        if (ni_logan_device_session_write(p_session_ctx, &ni_dec_ctx->seq_hdr_pkt, NI_LOGAN_DEVICE_TYPE_DECODER) < 0)
        {
          p_session_ctx->session_run_state = LOGAN_SESSION_RUN_STATE_NORMAL;
          return NI_LOGAN_RETCODE_FAILURE;
        }
      }
      return 0;
    }
  }
  return ret;
}

int ni_logan_decode_reset(ni_logan_dec_context_t *ni_dec_ctx)
{
  ni_logan_session_context_t *p_session_ctx = ni_dec_ctx->p_session_ctx;
  ni_logan_session_data_io_t *p_ni_packet = NULL;
  ni_logan_retcode_t ret = NI_LOGAN_RETCODE_FAILURE;
  ni_log(NI_LOG_TRACE, "NI XCoder decode reset\n");

  int ori_run_state = p_session_ctx->session_run_state;
  ret = ni_logan_device_session_close(p_session_ctx, ni_dec_ctx->eos, NI_LOGAN_DEVICE_TYPE_DECODER);

#ifdef _WIN32
  ni_logan_device_close(p_session_ctx->device_handle);
#elif __linux__
  ni_logan_device_close(p_session_ctx->device_handle);
  ni_logan_device_close(p_session_ctx->blk_io_handle);
#endif
  p_session_ctx->device_handle = NI_INVALID_DEVICE_HANDLE;
  p_session_ctx->blk_io_handle = NI_INVALID_DEVICE_HANDLE;

  if (ni_dec_ctx->input_data_fifo)
  {
    // push back
    do{
      p_ni_packet = ni_logan_fifo_generic_write(ni_dec_ctx->input_data_fifo);
    }while(p_ni_packet);
    // free all
    do{
      p_ni_packet = ni_logan_fifo_generic_read(ni_dec_ctx->input_data_fifo);
      if (p_ni_packet)
      {
        ni_logan_packet_buffer_free(&(p_ni_packet->data.packet));
      }
    }while(p_ni_packet);
  }
  ni_logan_fifo_free(ni_dec_ctx->input_data_fifo);
  ni_log(NI_LOG_DEBUG, "%s(): decoder frame fifo freed.\n", __FUNCTION__);

  ni_logan_device_session_context_clear(p_session_ctx);
  if (ni_logan_dec_reinit(ni_dec_ctx) < 0)
  {
    ni_log(NI_LOG_ERROR, "%s(): ni_logan_dec_reinit fail\n", __FUNCTION__);
    return NI_LOGAN_RETCODE_FAILURE;
  }
  p_session_ctx->session_run_state = ori_run_state;
  return 0;
}

void ni_logan_decoder_frame_free(ni_logan_session_data_io_t *p_session_data)
{
  ni_logan_decoder_frame_buffer_free(&p_session_data->data.frame);
}

void ni_logan_decode_flush_buffers(ni_logan_dec_context_t *ni_dec_ctx)
{
  ni_logan_session_context_t *p_session_ctx = ni_dec_ctx->p_session_ctx;
  ni_log(NI_LOG_DEBUG, "%s enter\n",__func__);
  ni_dec_ctx->dec_fifo_num = ni_dec_ctx->input_data_fifo->number_of_buffers_used;
  ni_dec_ctx->flush_buffer = NI_LOGAN_DECODE_FLUSH_STATUS_START;
}
