#pragma once
#ifdef __cplusplus
extern "C"
{
#endif
#include <errno.h>
#include <stddef.h>
#include "ni_defs_logan.h"
#include "ni_rsrc_api_logan.h"
#include "ni_util_logan.h"
#include "ni_device_api_logan.h"
#include "ni_av_codec_logan.h"

#define NI_LOGAN_NAL_VPS_BIT                  (0x01)
#define NI_LOGAN_NAL_SPS_BIT                  (0x01<<1)
#define NI_LOGAN_NAL_PPS_BIT                  (0x01<<2)
#define NI_LOGAN_GENERATE_ALL_NAL_HEADER_BIT  (0x01<<3)
/* enum for decode flush buffers status */
typedef enum
{
  NI_LOGAN_DECODE_FLUSH_STATUS_OFF = 0,
  NI_LOGAN_DECODE_FLUSH_STATUS_START = 1,
  NI_LOGAN_DECODE_FLUSH_STATUS_FLUSHING = 2,
};

typedef struct _ni_logan_dec_context_t {
  /* below are sync up from ffmpeg */
  char *dev_xcoder;         /* from the user command, which device allocation method we use */
  char *dev_xcoder_name;    /* dev name of the xcoder card to use */
  char *blk_xcoder_name;    /* blk name of the xcoder card to use */
  int dev_dec_idx;          /* index of the decoder on the xcoder card */
  char *dev_dec_name;       /* name of the decoder on the xcoder card */
  int keep_alive_timeout;   /* keep alive timeout setting */
  int set_high_priority;    /* set_high_priority */
  int ff_log_level;         /* ffmpeg log level */
  char *xcoder_opts;
  int enable_user_data_sei_passthru;
  int enable_check_packet;  // check source packet. Skip SEI payloads after VCL
  int custom_sei;
  int low_delay;
  int pkt_nal_bitmap;
  int fps_num;
  int fps_denom;
  long bit_rate;
  int width;
  int height;
  int codec_format;
  int bit_depth_factor;
  int pic_reorder_delay;

  int eos_pkt_received;
  int dec_wait; // in us
  
  void *p_session_ctx;
  void *p_param;
  ni_logan_fifo_buffer_t *input_data_fifo;
  ni_logan_packet_t seq_hdr_pkt;

  // stream header copied/saved from AVCodecContext.extradata
  int got_first_key_frame;
  uint8_t *extradata;
  int extradata_size;

  int64_t current_pts;
  /* offset of the cur. pkt relative to the start of the bit stream. used to */
  /* track the returned decoded frame and retrieve its saved pts */
  unsigned long long offset;

  int started;
  int draining;
  int flushing;
  int eos;
  int flush_buffer;
  int dec_fifo_num;
} ni_logan_dec_context_t;

/*!*****************************************************************************
 *  \brief  Initialize decode parameters and open decoder device 
 *          including ni_logan_dec_context_t,
 *                    ni_logan_session_context_t,
 *                    ni_logan_decoder_params_t
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_logan_dec_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 ******************************************************************************/
LIB_API int ni_logan_dec_init(ni_logan_dec_context_t *ni_dec_ctx);

/*!*****************************************************************************
 *  \brief  Initialize init device session_context and init default params
 *          including ni_logan_dec_context_t,
 *                    ni_logan_session_context_t,
 *                    ni_logan_decoder_params_t
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_logan_dec_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 ******************************************************************************/
LIB_API int ni_logan_setup_decoder(ni_logan_dec_context_t *ni_dec_ctx);


/*!*****************************************************************************
 *  \brief  Parse the packet and send the packet to decode device
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_logan_dec_context_t
 *
 *  \param[in/out] flag   flag=0 or 1 to do something.
 *  \param[in]  pkt_data  Pointer to input packet data
 *  \param[in]  pkt_size  packet size
 *  \param[in]  pkt_flags packet flags
 *  \param[in]  pkt_pts   packet pts
 *  \param[in]  pkt_dts   packet dts
 *  \param[in]  avctx_width avctx width
 *  \param[in]  avctx_height  avctx height
 *  \param[in]  extradata   Pointer to uint8_t
 *  \param[in]  extradata_size  extradata size
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 ******************************************************************************/
LIB_API int ni_logan_dec_send(ni_logan_dec_context_t *ni_dec_ctx,
                              uint8_t *pkt_data, int pkt_size, int pkt_flags,
                              long long pkt_pts, long long pkt_dts,
                              int avctx_width, int avctx_height,
                              uint8_t *extradata, int extradata_size);

/*!*****************************************************************************
 *  \brief  Decode is flushing or not, including ni_logan_dec_context_t,
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_logan_dec_context_t
 *
 *  \return On 1
 *                        HW dosn't do decode flushing
 *          On 0
 *                        HW is doing decode flushing
 ******************************************************************************/
LIB_API int ni_logan_dec_is_flushing(ni_logan_dec_context_t *ni_dec_ctx);

/*!*****************************************************************************
 *  \brief  Receive decoded data from decoder, including ni_logan_dec_context_t,
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_logan_dec_context_t
 *  \param[in/out] p_session_data  Pointer to ni_logan_session_data_io_t
 *  \param[in]  avctx_width avctx width
 *  \param[in]  avctx_height avctx height
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 ******************************************************************************/
LIB_API int ni_logan_dec_receive(ni_logan_dec_context_t *ni_dec_ctx,
                                    ni_logan_session_data_io_t *p_session_data,
                                    int avctx_width,
                                    int avctx_height);

/*!*****************************************************************************
 *  \brief  Decode close, including ni_logan_dec_context_t,
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_logan_dec_context_t
 *
 ******************************************************************************/
LIB_API void ni_logan_dec_close(ni_logan_dec_context_t *ni_dec_ctx);

/*!*****************************************************************************
 *  \brief  decoder frame buffer pool free, including void* 
 *
 *  \param[in]   opaque parameter to be got for processing or passed to free
 *  \param[in]   data  data array
 *
 ******************************************************************************/
LIB_API void ni_logan_buf_pool_free(void *opaque, uint8_t *data);

/*!*****************************************************************************
 *  \brief  decoder flush buffers
 *
 *  \param[in/out]   ni_dec_ctx    Pointer to ni_logan_dec_context_t
 *
 ******************************************************************************/
LIB_API void ni_logan_decode_flush_buffers(ni_logan_dec_context_t *ni_dec_ctx);

#ifdef __cplusplus
}
#endif