/*******************************************************************************
 *
 * Copyright (C) 2022 NETINT Technologies
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 ******************************************************************************/

/*!*****************************************************************************
*  \file   ni_enc_api_logan.h
*
*  \brief  NETINT encoder API header file
*
*******************************************************************************/
#pragma once
  
  
#ifdef __cplusplus
extern "C"
{
#endif

#include "ni_defs_logan.h"
#include "ni_rsrc_api_logan.h"
#include "ni_util_logan.h"
#include "ni_device_api_logan.h"
#include "ni_av_codec_logan.h"

typedef struct _ni_logan_enc_context
{
  char *dev_xcoder;         /* from the user command, which device allocation method we use */
  int  dev_enc_idx;         /* index of the decoder on the xcoder card */
  char *dev_enc_name;       /* dev name of the xcoder card to use */
  int  keep_alive_timeout;  /* keep alive timeout setting */
  int  set_high_priority;   /*set_high_priority*/

  int timebase_num;
  int timebase_den;
  int ticks_per_frame;
  int64_t bit_rate;
  int width;
  int height;
  int ff_log_level; /* ffmpeg log level */
  int codec_format;
  int pix_fmt;

  // color metrics
  int color_primaries;
  int color_trc;
  int color_space;
  int color_range;

  // sample_aspect_ratio
  int sar_num;
  int sar_den;

  void *p_session_ctx;  /* ni_logan_session_context_t */
  void *p_encoder_params; /* ni_logan_encoder_params_t */
  ni_logan_session_data_io_t *p_input_fme; /* used for sending raw data to xcoder */
  ni_logan_session_data_io_t  output_pkt; /* used for receiving bitstream from xcoder */
  ni_logan_fifo_buffer_t *input_data_fifo;

  /* Variables that need to be initialized */
  int started;
  uint8_t *p_spsPpsHdr;
  int spsPpsHdrLen;
  int spsPpsAttach; /* Attach SPS&PPS to Packet */
  int spsPpsArrived;
  int firstPktArrived; /* specially treat first pkt of the encoded bitstream */
  int dts_offset;
  int reconfigCount; /* count of reconfig */
  uint64_t total_frames_received;
  int64_t first_frame_pts;
  int64_t latest_dts;

  // original ConformanceWindowOffsets
  int orig_conf_win_top;   /*!*< A conformance window size of TOP */
  int orig_conf_win_bottom;   /*!*< A conformance window size of BOTTOM */
  int orig_conf_win_left;  /*!*< A conformance window size of LEFT */
  int orig_conf_win_right; /*!*< A conformance window size of RIGHT */

  // to generate encoded bitstream headers
  uint8_t *extradata;
  int extradata_size;

  // low delay mode flags
  int gotPacket; /* used to stop receiving packets when a packet is already received */
  int sentFrame; /* used to continue receiving packets when a frame is sent and a packet is not yet received */

  // sync framerate
  int fps_number;
  int fps_denominator;

  /* actual device index/name of encoder after opened */
  int  actual_dev_enc_idx;
  char *actual_dev_name;
  
  int eos_fme_received; // received the eos frame from the ffmpeg interface
  int encoder_flushing; // NI hardware encoder start flushing
  int encoder_eof; // recieved eof from NI hardware encoder
} ni_logan_enc_context_t;

/*!*****************************************************************************
 *  \brief  Initialize encode parameters, including ni_logan_enc_context_t,
 *          ni_logan_session_context_t and ni_logan_encoder_params_t
 *
 *  \param[in/out] p_enc_ctx  Pointer to ni_logan_enc_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 ******************************************************************************/
LIB_API int ni_logan_encode_init(ni_logan_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Parse encoder parameters
 *          Check whether the parameters are reasonable
 *
 *  \param[in/out] p_enc_ctx Pointer to ni_logan_enc_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE,
 *                        NI_LOGAN_RETCODE_INVALID_PARAM, etc.
 ******************************************************************************/
LIB_API int ni_logan_encode_params_parse(ni_logan_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Get encoder headers from the hardware
 *
 *  \param[in/out] p_enc_ctx  Pointer to ni_logan_enc_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 ******************************************************************************/
LIB_API int ni_logan_encode_header(ni_logan_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Open encoder
 *
 *  \param[in/out] p_enc_ctx Pointer to ni_logan_enc_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM, etc.
 ******************************************************************************/
LIB_API int ni_logan_encode_open(ni_logan_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Close encoder
 *
 *  \param[in/out] p_enc_ctx Pointer to ni_logan_enc_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC,
 *                        NI_LOGAN_RETCODE_ERROR_NVME_CMD_FAILED, etc
 ******************************************************************************/
LIB_API int ni_logan_encode_close(ni_logan_enc_context_t *p_enc_ctx);


/*!*****************************************************************************
 *  \brief  Get allocated ni_frame from fifo pool
 *
 *  \param[in]  p_enc_ctx Pointer to ni_logan_enc_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 ******************************************************************************/
LIB_API int ni_logan_encode_get_frame(ni_logan_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Fill reconfig vfr
 *
 *  \param[in]  p_enc_ctx Pointer to ni_logan_enc_context_t
 *  \param[in/out]  ni_frame aux_data may be updated
 *  \param[in]  pts frame pts
 *
 *  \return
 ******************************************************************************/
LIB_API void ni_logan_encode_reconfig_vfr(ni_logan_enc_context_t * p_enc_ctx,
                                          ni_logan_frame_t *ni_frame,
                                          int64_t pts);

/*!*****************************************************************************
 *  \brief  Copy frame data to NI frame
 *
 *  \param[in/out] p_enc_ctx Pointer to ni_logan_enc_context_t
 *  \param[in] p_dec_frame   Pointer to ni_logan_frame_t,
 *                           contains  aux data from NI decoder
 *  \param[in] p_data        Pointer to input frame data
 *  \param[in] linesize      Pointer to linesize of input frame
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 ******************************************************************************/
LIB_API int ni_logan_encode_copy_frame_data(ni_logan_enc_context_t *p_enc_ctx,
                                            ni_logan_frame_t *p_dec_frame,
                                            uint8_t *p_data[NI_LOGAN_MAX_NUM_DATA_POINTERS],
                                            int linesize[NI_LOGAN_MAX_NUM_DATA_POINTERS]);


/*!*****************************************************************************
 *  \brief  Send encode data to NI device
 *
 *  \param[in] p_enc_ctx  Pointer to ni_logan_enc_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE,
 *                        NI_LOGAN_RETCODE_INVALID_PARAM, etc.
 ******************************************************************************/
LIB_API int ni_logan_encode_send(ni_logan_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Copy NI packet to packet data buffer
 *
 *  \param[in] p_enc_ctx    Pointer to ni_logan_enc_context_t
 *  \param[out] p_data      Pointer to output packet data buffer,
 *  \param[in] first_packet flag of first packet
 *  \param[in] sps_pps_attach  flag of attached SPS/PPS
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE,
 *                        NI_LOGAN_RETCODE_PARAM_ERROR_WIDTH_TOO_BIG, etc
 ******************************************************************************/
LIB_API int ni_logan_encode_copy_packet_data(ni_logan_enc_context_t *p_enc_ctx,
                                             uint8_t *p_data,
                                             int first_packet,
                                             int sps_pps_attach);

/*!*****************************************************************************
 *  \brief  Receive encoded data from NI device
 *
 *  \param[in] p_enc_ctx  Pointer to ni_logan_enc_context_t
 *
 *  \return 0 when no packets received
 *          > 0 when new packet received, means the size of packet
 *          < 0 when failed, NI_LOGAN_RETCODE_FAILURE,
 *                           NI_LOGAN_RETCODE_INVALID_PARAM, etc.
 ******************************************************************************/
LIB_API int ni_logan_encode_receive(ni_logan_enc_context_t *p_enc_ctx);

#ifdef __cplusplus
}
#endif
