/*******************************************************************************
 *
 * Copyright (C) 2022 NETINT Technologies
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 ******************************************************************************/

/*!*****************************************************************************
 *
 *   \file          ni_rsrc_update_logan.c
 *
 *   @date          May 10, 2019
 *
 *   \brief         NETINT T408 resource update utility program
 *
 *   @author
 *
 ******************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#if defined(__linux__) || defined(__APPLE__)
#include <unistd.h>
#include <sys/types.h>
#endif

#include "ni_defs_logan.h"
#include "ni_rsrc_api_logan.h"
#include "ni_rsrc_priv_logan.h"
#include "ni_util_logan.h"

#ifdef _WIN32
#include "ni_getopt_logan.h"
#define DEV_NAME_PREFIX "\\\\.\\Scsi"
#elif defined(__linux__)
#define DEV_NAME_PREFIX "/dev/nvme"
#define NVME_PCI_PREFIX "nvme/nvme"
#define NVME_SUBSYS_PREFIX "nvme-subsystem/nvme-subsys"
#elif defined(__APPLE__)
#define DEV_NAME_PREFIX "/dev/disk"
#endif

/*!******************************************************************************
 *  \brief  get the NVMe device's character device name (e.g. /dev/nvmeX)
 *
 *  \param  p_in_buf    whole device name passed in
 *          p_out_buf   full device name without name-space/partition-number
 *          out_buf_len maximum size of output buffer
 *
 *  \return    0 if value device name is found, -1 otherwise
 ******************************************************************************/
static int get_device_name(const char *p_in_buf, char *p_out_buf, int out_buf_len)
{
  FILE *cmd_fp;
  char cmd[128] = {'\0'};
  char cmd_ret[128] = {'\0'};
  const char *p_src;
  char *p_dst;

  if (!p_in_buf || !p_out_buf)
  {
    return -1;
  }

  p_src = p_in_buf;
  p_dst = p_out_buf;
#ifdef _WIN32
  ni_log(NI_LOG_TRACE, "In Windows block name equals to device name\n");
  snprintf(p_out_buf, out_buf_len, "%s", p_in_buf);
  return 0;
#elif defined(XCODER_LINUX_VIRTIO_DRIVER_ENABLED)
  ni_log(NI_LOG_TRACE, "The block path is same as device path in Linux virtual machine with VirtIO driver.\n");
  snprintf(p_out_buf, out_buf_len, "%s", p_in_buf);
  return 0;
#elif defined(__APPLE__)
  snprintf(p_out_buf, out_buf_len, "%s", p_in_buf);
  return 0;
#else
#ifdef __ANDROID__
  ni_log(NI_LOG_TRACE, "For Android, assume index block namespace is same as device index\n");
  p_src = strstr(p_in_buf, "/nvme");
  if (p_src != NULL)
  {
    int len = strlen("/nvme");
    strcpy(p_out_buf, p_in_buf);
    p_src += len;
    p_dst += len;
    while (isdigit(*p_src))
    {
      *p_dst++ = *p_src++;
    }
    *p_dst = '\0';
    return 0;
  }
#else
  // Note, we are using udevadm through terminal instead of libudev.c to avoid requiring extra kernel dev packges
  snprintf(cmd, sizeof(cmd) - 1, "udevadm info -q path -n %s", p_in_buf);

  // check p_in_buf exists in /dev/ folder. If not, return NI_LOGAN_RETCODE_FAILURE
  if (access(p_in_buf, F_OK) == -1)
  {
    return NI_LOGAN_RETCODE_FAILURE;
  }

  // look for child block in sysfs mapping tree
  cmd_fp = popen(cmd, "r");
  if (!cmd_fp)
  {
    return NI_LOGAN_RETCODE_FAILURE;
  }

  if (fgets(cmd_ret, sizeof(cmd_ret)/sizeof(cmd_ret[0]), cmd_fp) != NULL)
  {
    p_src = strstr(cmd_ret, NVME_PCI_PREFIX);
    if (p_src != NULL)
    {
      p_src += strlen(NVME_PCI_PREFIX);
    }
    else
    {
      p_src = strstr(cmd_ret, NVME_SUBSYS_PREFIX);
      if (p_src != NULL)
      {
        p_src += strlen(NVME_SUBSYS_PREFIX);
      }
      else
      {
        // Not found.
        ni_log(NI_LOG_ERROR, "Failed to parse NVMe udev path %s.\n", cmd_ret);
        pclose(cmd_fp);
        return -1;
      }
    }
    strcpy(p_out_buf, DEV_NAME_PREFIX);
    p_dst += strlen(DEV_NAME_PREFIX);
    while (isdigit(*p_src))
    {
      *p_dst++ = *p_src++;
    }
    *p_dst = '\0';
    pclose(cmd_fp);
    return 0;
  }
#endif
  else
  {
    ni_log(NI_LOG_TRACE, "Failed to find device name.\n");
    pclose(cmd_fp);
    return -1;
  }
#endif
}

static void display_help(void)
{
  printf("-------- ni_rsrc_update_logan v%s --------\n"
         "Update NetInt Logan transcoder resource (encoders and decoders) status.\n"
         "-a <nvme_device>    Create a resource entry for a newly active transcoder. (eg. /dev/nvme0) \n"
         "-d <nvme_device>    Delete the resource entry for a removed transcoder card. (eg. /dev/nvme0)\n"
         "-r    Init transcoder card resource regardless firmware release version to \n"
         "      libxcoder_logan version compatibility. Must be used with -a option.\n"
         "      Default: only init cards with compatible firmware version.\n"
         "-c    Only init if all cards are fully compatible. Must be used with -a option.\n"
         "      Default: init as many fully and partially cards as possible.\n"
         "-l    Set loglevel of libxcoder_logan API.\n"
         "      [none, fatal, error, info, debug, trace]\n"
         "      Default: info\n"
         "-v    Show libxcoder_logan version.\n"
         "-h    Display this help and exit.\n", NI_LOGAN_XCODER_REVISION);
}

/*!******************************************************************************
 *  \brief
 *
 *  \param
 *
 *  \return
 ******************************************************************************/
int main(int argc, char *argv[])
{
  int opt, rc = 0;
  char char_dev_name[64];
  int should_match_rev = 1;
  int init_no_compat_check = 0;
  int init_only_if_full_compat = 0;
  int add_dev = 0;
  int delete_dev = 0;

  if (argc == 1) {
    display_help();
    return 0;
  }

  // arg handling
  while ((opt = getopt(argc, argv, "hvrca:d:l:")) != -1) {
    switch (opt) {
    case 'a':
#ifdef __linux__
      rc = get_device_name(optarg, char_dev_name, sizeof(char_dev_name));
      add_dev = 1;
#endif
      break;
    case 'd':
#ifdef __linux__
      rc = get_device_name(optarg, char_dev_name, sizeof(char_dev_name));
      delete_dev = 1;
#endif
      break;
    case 'r':
      init_no_compat_check = 1;
      break;
    case 'c':
      init_only_if_full_compat = 1;
      break;
    case 'l':
      if (!strcmp(optarg, "none")) {
        ni_log_set_level(NI_LOG_NONE);
      } else if (!strcmp(optarg, "fatal")) {
        ni_log_set_level(NI_LOG_FATAL);
      } else if (!strcmp(optarg, "error")) {
        ni_log_set_level(NI_LOG_ERROR);
      } else if (!strcmp(optarg, "info")) {
        ni_log_set_level(NI_LOG_INFO);
      } else if (!strcmp(optarg, "debug")) {
        ni_log_set_level(NI_LOG_DEBUG);
      } else if (!strcmp(optarg, "trace")) {
        ni_log_set_level(NI_LOG_TRACE);
      } else {
        fprintf(stderr, "unknown log level selected: %s", optarg);
        return 1;
      }
      break;
    case 'v':
      printf("%s\n", NI_LOGAN_XCODER_REVISION);
      return 0;
    case 'h':
    default:
      display_help();
      return 0;
    }
  }

  // check option
  if (add_dev && delete_dev)
  {
    fprintf(stderr, "Error: can not add and delete device at the same time\n\n");
    display_help();
    return 1;
  }

  if (init_no_compat_check && init_only_if_full_compat)
  {
    fprintf(stderr, "Error: -r option cannot be used with -c option\n\n");
    display_help();
    return 1;
  }

  if ((init_no_compat_check || init_only_if_full_compat) && !add_dev)
  {
    fprintf(stderr, "Error: %s option must be used with -a option\n\n",
            init_no_compat_check ? "-r" : "-c");
    display_help();
    return 1;
  }

  // process options
  if (init_no_compat_check)
  {
    should_match_rev = 0;
  }
  else if (init_only_if_full_compat)
  {
    should_match_rev = 2;
  }

  if (add_dev)
  {
    if (rc || (rc = ni_logan_rsrc_add_device(char_dev_name, should_match_rev)))
      printf("%s not added as transcoder.\n", optarg);
    else
      printf("Added transcoder %s successfully.\n", char_dev_name);
    return rc;
  }
  else if (delete_dev)
  {
    if (rc || (rc = ni_logan_rsrc_remove_device(char_dev_name)))
      printf("%s not removed as transcoder.\n", optarg);
    else
      printf("Removed transcoder %s successfully.\n", char_dev_name);
    return rc;
  }
}
