/*******************************************************************************
 *
 * Copyright (C) 2022 NETINT Technologies
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 ******************************************************************************/

/*!*****************************************************************************
*  \file   ni_util_logan.c
*
*  \brief  Exported utility routines
*
*******************************************************************************/

#if defined(__linux__) || defined(__APPLE__)
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#endif

#include "ni_nvme_logan.h"
#include "ni_util_logan.h"

#ifdef XCODER_SIM_ENABLED
static int32_t sim_run_count = 0;
static int32_t sim_eos_flag = 0;
#endif

/*!
 * timezone information is stored outside the kernel so tzp isn't used anymore.
 *
 * Note: this function is not for Win32 high precision timing purpose. See
 * elapsed_time().
 */
int32_t ni_logan_gettimeofday(struct timeval *p_tp, void *p_tzp)
{
#ifdef _WIN32
  FILETIME    file_time;
  SYSTEMTIME  system_time;
  ULARGE_INTEGER ularge;
  /*! FILETIME of Jan 1 1970 00:00:00. */
  static const unsigned __int64 epoch = ((unsigned __int64)116444736000000000ULL);

  (void)p_tzp; /*!Remove compiler warnings*/

  GetSystemTime(&system_time);
  SystemTimeToFileTime(&system_time, &file_time);
  ularge.LowPart = file_time.dwLowDateTime;
  ularge.HighPart = file_time.dwHighDateTime;

  p_tp->tv_sec = (long)((ularge.QuadPart - epoch) / 10000000L);
  p_tp->tv_usec = (long)(system_time.wMilliseconds * 1000);

  return 0;
#else
  return gettimeofday(p_tp, p_tzp);
#endif
}

uint64_t ni_logan_gettime_ns(void)
{
#ifdef _WIN32
  LARGE_INTEGER frequency;
  LARGE_INTEGER count;
  uint64_t time_sec, time_nsec;

  // Get frequency firstly
  QueryPerformanceFrequency(&frequency);

  QueryPerformanceCounter(&count);

  time_sec = count.QuadPart / frequency.QuadPart;
  time_nsec = (count.QuadPart - time_sec * frequency.QuadPart) * 1000000000LL / frequency.QuadPart;

  return (time_sec * 1000000000LL + time_nsec);
#else
  struct timespec ts;
  clock_gettime(CLOCK_REALTIME, &ts);
  return ts.tv_sec * 1000000000LL + ts.tv_nsec;
#endif
}

uint32_t ni_logan_round_up(uint32_t  number_to_round, uint32_t  multiple)
{
  uint32_t remainder;

  if (0 == multiple)
  {
    return number_to_round;
  }

  remainder = number_to_round % multiple;
  if (0 == remainder)
  {
    return number_to_round;
  }

  return number_to_round + multiple - remainder;
}

int32_t ni_logan_posix_memalign(void ** pp_memptr, size_t alignment, size_t size)
{
#ifdef _WIN32
  *pp_memptr = malloc(size);
  if (NULL == *pp_memptr)
  {
    return 1;
  }
  else
  {
    ZeroMemory(*pp_memptr, size);
    return 0;
  }
#else
  return posix_memalign(pp_memptr, alignment, size);
#endif
}

#ifdef __linux__
static inline int ni_logan_min(int a, int b)
{
  return a < b ? a : b;
}

static inline int ni_logan_max(int a, int b)
{
  return a > b ? a : b;
}

/*!******************************************************************************
*  \brief  Get max io transfer size from the kernel
*
*  \param
*
*  \return
*******************************************************************************/
uint32_t ni_logan_get_kernel_max_io_size(const char * p_dev)
{
  FILE *p_file = NULL;   /* file pointer*/
  char file_name[KERNEL_NVME_FILE_NAME_MAX_SZ];
  uint32_t max_segments = 0, min_io_size = 0, max_hw_sectors_kb = 0;
  uint32_t io_size = DEFAULT_IO_TRANSFER_SIZE;
  int len = 0, err = 0;


  if ( !p_dev )
  {
    ni_log(NI_LOG_TRACE, "Invalid Arguments\n");
    LRETURN;
  }

  len = strlen(p_dev) - 5;
  if(len < MIN_NVME_BLK_NAME_LEN)
  {
    ni_log(NI_LOG_TRACE, "p_dev length is %d\n",len);
    LRETURN;
  }

  // Get Max number of segments from /sys
  memset(file_name,0,sizeof(file_name));
  memcpy(file_name,SYS_PARAMS_PREFIX_PATH,SYS_PREFIX_SZ);
  //start from 5 chars ahead to not copy the "/dev/" since we only need whats after it
  if (strstr(p_dev, "block"))
  {
    strncat(file_name, (char*)(p_dev + 11) , sizeof(file_name) - SYS_PREFIX_SZ);
  }
  else
  {
    strncat(file_name, (char*)(p_dev + 5) , sizeof(file_name) - SYS_PREFIX_SZ);
  }
  strncat(file_name, KERNEL_NVME_MAX_SEG_PATH, sizeof(file_name) - SYS_PREFIX_SZ - len);
  ni_log(NI_LOG_TRACE, "file_name is %s\n",file_name);
  p_file = fopen(file_name,"r");
  if (!p_file)
  {
    ni_log(NI_LOG_ERROR, "Error %d: file_name failed to open: %s\n",
    NI_ERRNO, file_name);
    LRETURN;
  }

  err = fscanf(p_file,"%u",&max_segments);
  if (EOF == err)
  {
    ni_log(NI_LOG_ERROR, "Error %d: fscanf failed on: %s max_segments\n",
    NI_ERRNO, file_name);
    LRETURN;
  }

  fclose(p_file);
  p_file = NULL;
  // Get Max segment size from /sys
  memset(file_name,0,sizeof(file_name));
  memcpy(file_name,SYS_PARAMS_PREFIX_PATH,SYS_PREFIX_SZ);
  if (strstr(p_dev, "block"))
  {
    strncat(file_name, (char*)(p_dev + 11) , sizeof(file_name) - SYS_PREFIX_SZ);
  }
  else
  {
    strncat(file_name, (char*)(p_dev + 5) , sizeof(file_name) - SYS_PREFIX_SZ);
  }
  strncat(file_name, KERNEL_NVME_MIN_IO_SZ_PATH, sizeof(file_name) - SYS_PREFIX_SZ - len);
  ni_log(NI_LOG_TRACE, "file_name  is %s\n",file_name);
  p_file = fopen(file_name,"r");
  if (!p_file)
  {
    ni_log(NI_LOG_ERROR, "Error %d: file_name  failed to open: %s\n",
    NI_ERRNO, file_name);
    LRETURN;
  }

  err = fscanf(p_file,"%u",&min_io_size);
  if (EOF == err)
  {
    ni_log(NI_LOG_TRACE, "fscanf failed on: %s min_io_size\n",file_name);
    LRETURN;
  }

  fclose(p_file);
  p_file = NULL;
  //Now get max_hw_sectors_kb
  memset(file_name,0,sizeof(file_name));
  memcpy(file_name,SYS_PARAMS_PREFIX_PATH,SYS_PREFIX_SZ);
  if (strstr(p_dev, "block"))
  {
    strncat(file_name, (char*)(p_dev + 11) , sizeof(file_name) - SYS_PREFIX_SZ);
  }
  else
  {
    strncat(file_name, (char*)(p_dev + 5) , sizeof(file_name) - SYS_PREFIX_SZ);
  }
  strncat(file_name, KERNEL_NVME_MAX_HW_SEC_KB_PATH, sizeof(file_name) - SYS_PREFIX_SZ - len);
  ni_log(NI_LOG_TRACE, "file_name  is %s\n",file_name);
  p_file = fopen(file_name,"r");
  if (!p_file)
  {
    ni_log(NI_LOG_ERROR, "Error %d: file_name  failed to open: %s\n",
           NI_ERRNO, file_name);
    LRETURN;
  }

  err = fscanf(p_file,"%u",&max_hw_sectors_kb);
  if (EOF == err)
  {
    ni_log(NI_LOG_TRACE, "fscanf failed on: %s min_io_size\n",file_name);
    LRETURN;
  }

  if( ni_logan_min(min_io_size * max_segments, max_hw_sectors_kb * 1024) > MAX_IO_TRANSFER_SIZE )
  {
    io_size = MAX_IO_TRANSFER_SIZE;
  }
  else
  {
    io_size = ni_logan_min( min_io_size * max_segments, max_hw_sectors_kb * 1024);
  }

  ni_log(NI_LOG_DEBUG, "\nMAX NVMe IO Size of %d was calculated for this platform "
         "and will be used unless overwritten by user settings\n", io_size);
  fflush(stderr);

  END:

  if(p_file)
  {
    fclose(p_file);
  }

  return io_size;
}

#endif

void ni_logan_usleep(int64_t usec)
{
#ifdef _WIN32
  if(usec < 5000)  //this will be more accurate when less than 5000
  {
    LARGE_INTEGER Count;
    LARGE_INTEGER Count_;
    LARGE_INTEGER Frequency;
    QueryPerformanceCounter(&Count);
    QueryPerformanceFrequency(&Frequency);
    Count_.QuadPart = Count.QuadPart + usec*(Frequency.QuadPart/1000000);
    do
    {
      QueryPerformanceCounter(&Count);
    } while(Count.QuadPart<Count_.QuadPart);
  }
  else
  {
    HANDLE timer = NULL;
    LARGE_INTEGER ft = { 0 };
    BOOL retval = FALSE;

    ft.QuadPart = -(10 * usec); // Convert to 100 nanosecond interval, negative value indicates relative time

    timer = CreateWaitableTimer(NULL, TRUE, NULL);
    if (NULL != timer)
    {
      retval = SetWaitableTimer(timer, &ft, 0, NULL, NULL, 0);
      if (retval)
      {
        WaitForSingleObject(timer, INFINITE);
      }
    }

    if (timer)
    {
      CloseHandle(timer);
    }
  }
#else
  if (usec < 0xFFFFFFFF)  // to avoid overflow
  {
    usleep(usec);
  }
  else
  {
    sleep(usec >> 32);
  }
#endif
}

// memory buffer pool operations (one use is for decoder frame buffer pool)
// expand buffer pool by a pre-defined size
ni_logan_buf_t *ni_logan_buf_pool_expand(ni_logan_buf_pool_t *pool)
{
  int32_t i;
  for (i = 0; i < NI_LOGAN_DEC_FRAME_BUF_POOL_SIZE_EXPAND; i++)
  {
    if( NULL == ni_logan_buf_pool_allocate_buffer(pool, pool->buf_size) )
    {
      ni_log(NI_LOG_ERROR, "FATAL ERROR: Failed to expand allocate pool buffer for "
             "pool :%p  current size: %d\n", pool, pool->number_of_buffers);
      return NULL;
    }
  }
  pool->number_of_buffers += NI_LOGAN_DEC_FRAME_BUF_POOL_SIZE_EXPAND;
  return pool->p_free_head;
}

// get a free memory buffer from the pool
ni_logan_buf_t *ni_logan_buf_pool_get_buffer(ni_logan_buf_pool_t *p_buffer_pool)
{
  ni_logan_buf_t *buf = NULL;

  if (NULL == p_buffer_pool)
  {
    return NULL;
  }

  ni_logan_pthread_mutex_lock(&(p_buffer_pool->mutex));
  buf = p_buffer_pool->p_free_head;

  // find and return a free buffer
  if (NULL == buf)
  {
    ni_log(NI_LOG_ERROR, "Expanding dec fme buffer_pool from %d to %d \n",
           p_buffer_pool->number_of_buffers,
           p_buffer_pool->number_of_buffers+NI_LOGAN_DEC_FRAME_BUF_POOL_SIZE_EXPAND);

    buf = ni_logan_buf_pool_expand(p_buffer_pool);

    if (NULL == buf)
    {
      ni_logan_pthread_mutex_unlock(&(p_buffer_pool->mutex));
      return NULL;
    }
  }

  // remove it from free list head; reconnect the linked list, the p_next
  // will become the new head now
  p_buffer_pool->p_free_head = buf->p_next_buffer;

  if ( NULL != buf->p_next_buffer )
  {
    buf->p_next_buffer->p_previous_buffer = NULL;
  }
  else
  {
    p_buffer_pool->p_free_tail = NULL;
  }

  // add it to the used list tail
  buf->p_previous_buffer = p_buffer_pool->p_used_tail;
  buf->p_next_buffer = NULL;

  if ( NULL != p_buffer_pool->p_used_tail )
  {
    p_buffer_pool->p_used_tail->p_next_buffer = buf;
  }
  else
  {
    p_buffer_pool->p_used_head = buf;
  }

  p_buffer_pool->p_used_tail = buf;

  ni_logan_pthread_mutex_unlock(&(p_buffer_pool->mutex));

  ni_log(NI_LOG_TRACE, "%s ptr %p  buf %p\n", __FUNCTION__, buf->buf, buf);
  return buf;
}

// return a used memory buffer to the pool
void ni_logan_buf_pool_return_buffer(ni_logan_buf_t *buf, ni_logan_buf_pool_t *p_buffer_pool)
{
  // p_buffer_pool could be null in case of delayed buffer return after pool
  // has been freed
  if(! buf)
  {
    return;
  }

  ni_log(NI_LOG_TRACE, "%s ptr %p  buf %p\n", __FUNCTION__, buf->buf, buf);

  if (! p_buffer_pool)
  {
    ni_log(NI_LOG_TRACE, "%s: pool already freed self destroy\n", __FUNCTION__);
    free(buf->buf);
    free(buf);
    return;
  }

  ni_logan_pthread_mutex_lock(&(p_buffer_pool->mutex));

  // remove buf from the used list
  if (NULL != buf->p_previous_buffer)
  {
    buf->p_previous_buffer->p_next_buffer = buf->p_next_buffer;
  }
  else
  {
    p_buffer_pool->p_used_head = buf->p_next_buffer;
  }

  if (NULL != buf->p_next_buffer)
  {
    buf->p_next_buffer->p_previous_buffer = buf->p_previous_buffer;
  }
  else
  {
    p_buffer_pool->p_used_tail = buf->p_previous_buffer;
  }

  // put it on the tail of free buffers list
  buf->p_previous_buffer = p_buffer_pool->p_free_tail;
  buf->p_next_buffer = NULL;

  if (NULL != p_buffer_pool->p_free_tail)
  {
    p_buffer_pool->p_free_tail->p_next_buffer = buf;
  }
  else
  {
    p_buffer_pool->p_free_head = buf;
  }

  p_buffer_pool->p_free_tail = buf;

  // shrink the buffer pool by EXPAND size if free list size is at least
  // 2x EXPAND to manage buffer pool expansion
  int count = 0;
  ni_logan_buf_t *p_tmp = p_buffer_pool->p_free_head;
  while (p_tmp)
  {
    count++;
    p_tmp = p_tmp->p_next_buffer;
  }

  if (count >= 2 * NI_LOGAN_DEC_FRAME_BUF_POOL_SIZE_EXPAND)
  {
    ni_log(NI_LOG_INFO, "%s shrink buf pool free size from %d by %d, "
           "current total: %u\n",
           __FUNCTION__, count, NI_LOGAN_DEC_FRAME_BUF_POOL_SIZE_EXPAND,
           p_buffer_pool->number_of_buffers);

    p_tmp = p_buffer_pool->p_free_head;
    while (p_tmp && count > NI_LOGAN_DEC_FRAME_BUF_POOL_SIZE_EXPAND)
    {
      p_buffer_pool->p_free_head = p_tmp->p_next_buffer;
      free(p_tmp->buf);
      free(p_tmp);
      p_tmp = p_buffer_pool->p_free_head;
      count--;
      p_buffer_pool->number_of_buffers--;
    }
  }

  ni_logan_pthread_mutex_unlock(&(p_buffer_pool->mutex));
}

// allocate a memory buffer and place it in the pool
ni_logan_buf_t *ni_logan_buf_pool_allocate_buffer(ni_logan_buf_pool_t *p_buffer_pool,
                                      int buffer_size)
{
  ni_logan_buf_t *p_buffer = NULL;
  void *p_buf = NULL;

  if (NULL != p_buffer_pool &&
     (p_buffer = (ni_logan_buf_t *)malloc(sizeof(ni_logan_buf_t))) != NULL)
  {
    // init the struct
    memset(p_buffer, 0, sizeof(ni_logan_buf_t));

    if (ni_logan_posix_memalign(&p_buf, sysconf(_SC_PAGESIZE), buffer_size))
    {
      free(p_buffer);
      ni_log(NI_LOG_ERROR, "ERROR %d: %s() failed\n",
             NI_ERRNO, __FUNCTION__);
      return NULL;
    }
    ni_log(NI_LOG_TRACE, "%s ptr %p  buf %p\n", __FUNCTION__, p_buf, p_buffer);
    p_buffer->buf = p_buf;
    p_buffer->pool = p_buffer_pool;

    // add buffer to the buf pool list
    p_buffer->p_prev = NULL;
    p_buffer->p_next = NULL;
    p_buffer->p_previous_buffer = p_buffer_pool->p_free_tail;

    if (p_buffer_pool->p_free_tail != NULL)
    {
      p_buffer_pool->p_free_tail->p_next_buffer = p_buffer;
    }
    else
    {
      p_buffer_pool->p_free_head = p_buffer;
    }

    p_buffer_pool->p_free_tail = p_buffer;
  }

  return p_buffer;
}

// decoder frame buffer pool init & free
int32_t ni_logan_dec_fme_buffer_pool_initialize(ni_logan_session_context_t* p_ctx,
                                          int32_t number_of_buffers,
                                          int width,
                                          int height,
                                          int height_align,
                                          int factor)
{
  int32_t i;
  int width_aligned = width;
  int height_aligned = height;

  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  width_aligned = ((width + 31) / 32) * 32;
  height_aligned = ((height + 7) / 8) * 8;
  if (height_align)
  {
    height_aligned = ((height + 15) / 16) * 16;
  }

  uint32_t luma_size = width_aligned * height_aligned * factor;
  uint32_t chroma_b_size = luma_size / 4;
  uint32_t chroma_r_size = chroma_b_size;
  uint32_t buffer_size = luma_size + chroma_b_size + chroma_r_size +
                         NI_LOGAN_FW_META_DATA_SZ + NI_LOGAN_MAX_SEI_DATA;

  // added 2 blocks of 512 bytes buffer space to handle any extra metadata
  // retrieval from fw
  buffer_size = ((buffer_size + (NI_LOGAN_MEM_PAGE_ALIGNMENT - 1)) / NI_LOGAN_MEM_PAGE_ALIGNMENT) *
                NI_LOGAN_MEM_PAGE_ALIGNMENT + NI_LOGAN_MEM_PAGE_ALIGNMENT * 3;

  if (p_ctx->dec_fme_buf_pool != NULL)
  {
    ni_log(NI_LOG_TRACE, "Warning init dec_fme Buf pool already with size %d\n",
           p_ctx->dec_fme_buf_pool->number_of_buffers);

    if (buffer_size > p_ctx->dec_fme_buf_pool->buf_size)
    {
      ni_log(NI_LOG_ERROR, "Warning resolution %dx%d memory buffer size %d > %d "
             "(existing buffer size), re-allocating !\n", width, height,
             buffer_size, p_ctx->dec_fme_buf_pool->buf_size);
      ni_logan_dec_fme_buffer_pool_free(p_ctx->dec_fme_buf_pool);
    }
    else
    {
      ni_log(NI_LOG_ERROR, "INFO resolution %dx%d memory buffer size %d <= %d "
             "(existing buffer size), continue !\n", width, height,
             buffer_size, p_ctx->dec_fme_buf_pool->buf_size);
      return 0;
    }
  }

  if ((p_ctx->dec_fme_buf_pool = (ni_logan_buf_pool_t *) malloc(sizeof(ni_logan_buf_pool_t))) == NULL)
  {
    ni_log(NI_LOG_ERROR, "Error %d: alloc for dec fme buf pool\n",
           NI_ERRNO);
    return -1;
  }

  // init the struct
  memset(p_ctx->dec_fme_buf_pool, 0, sizeof(ni_logan_buf_pool_t));
  ni_logan_pthread_mutex_init(&(p_ctx->dec_fme_buf_pool->mutex));
  p_ctx->dec_fme_buf_pool->number_of_buffers = number_of_buffers;

  ni_log(NI_LOG_TRACE, "%s: entries %d  entry size %d\n",
         __FUNCTION__, number_of_buffers, buffer_size);

  p_ctx->dec_fme_buf_pool->buf_size = buffer_size;
  for (i = 0; i < number_of_buffers; i++)
  {
    if (NULL == ni_logan_buf_pool_allocate_buffer(p_ctx->dec_fme_buf_pool, buffer_size))
    {
      // release everything we have allocated so far and exit
      ni_logan_dec_fme_buffer_pool_free(p_ctx->dec_fme_buf_pool);
      return -1;
    }
  }

  ni_log(NI_LOG_TRACE, "%s: exit\n", __FUNCTION__);
  return 0;
}

void ni_logan_dec_fme_buffer_pool_free(ni_logan_buf_pool_t *p_buffer_pool)
{
  ni_logan_buf_t *buf, *p_next;
  int32_t count_free = 0;

  ni_log(NI_LOG_TRACE, "%s: enter.\n", __FUNCTION__);

  if (p_buffer_pool)
  {
    // mark used buf not returned at pool free time by setting pool ptr in used
    // buf to NULL, so they will self-destroy when time is due eventually
    ni_logan_pthread_mutex_lock(&(p_buffer_pool->mutex));

    if (buf = p_buffer_pool->p_used_head)
    {
      p_next = buf;
      while (buf)
      {
        p_next = buf->p_next_buffer;
        ni_log(NI_LOG_TRACE, "Release ownership of ptr %p buf %p\n", buf->buf, buf);
        buf->pool = NULL;
        buf = p_next;
      }
    }
    ni_logan_pthread_mutex_unlock(&(p_buffer_pool->mutex));

    buf = p_buffer_pool->p_free_head;
    p_next = buf;
    // free all the buffers in the free list
    while (buf)
    {
      p_next = buf->p_next_buffer;
      free(buf->buf);
      free(buf);
      buf = p_next;
      count_free++;
    }

    if (count_free != p_buffer_pool->number_of_buffers)
    {
      ni_log(NI_LOG_TRACE, "%s free %d != number_of_buffers %d\n",
             __FUNCTION__, count_free, p_buffer_pool->number_of_buffers);
    }
    else
    {
      ni_log(NI_LOG_TRACE, "%s all buffers freed: %d.\n",
             __FUNCTION__, count_free);
    }
    free(p_buffer_pool);
    p_buffer_pool = NULL;
  }
  else
  {
    ni_log(NI_LOG_INFO, "%s: NOT allocated\n", __FUNCTION__);
  }
}

void ni_logan_buffer_pool_free(ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  ni_logan_queue_node_t *buf, *p_next;
  int32_t count = 0;

  ni_log(NI_LOG_TRACE, "%s: enter.\n", __FUNCTION__);

  if (p_buffer_pool)
  {
    buf = p_buffer_pool->p_free_head;
    p_next = buf;
    // free all the buffers in the free and used list
    while (buf)
    {
      p_next = buf->p_next_buffer;
      free(buf);
      buf = p_next;
      count++;
    }

    buf = p_buffer_pool->p_used_head;
    p_next = buf;
    while (buf)
    {
      p_next = buf->p_next_buffer;
      free(buf);
      buf = p_next;
      count++;
    }

    if (count != p_buffer_pool->number_of_buffers)
    {
      ni_log(NI_LOG_ERROR, "??? freed %d != number_of_buffers %d\n",
             count, p_buffer_pool->number_of_buffers);
    }
    else
    {
      ni_log(NI_LOG_TRACE, "p_buffer_pool freed %d buffers.\n", count);
    }
    free(p_buffer_pool);
    p_buffer_pool = NULL;
  }
  else
  {
    ni_log(NI_LOG_ERROR, "%s: NOT allocated\n", __FUNCTION__);
  }
}

ni_logan_queue_node_t *ni_logan_buffer_pool_allocate_buffer(ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  ni_logan_queue_node_t *p_buffer = NULL;

  if (NULL != p_buffer_pool && ( (p_buffer = (ni_logan_queue_node_t *)malloc(sizeof(ni_logan_queue_node_t))) != NULL) )
  {
    //Inititalise the struct
    memset(p_buffer,0,sizeof(ni_logan_queue_node_t));
    // add buffer to the buf pool list
    p_buffer->p_prev = NULL;
    p_buffer->p_next = NULL;
    p_buffer->p_previous_buffer = p_buffer_pool->p_free_tail;

    if (p_buffer_pool->p_free_tail != NULL)
    {
      p_buffer_pool->p_free_tail->p_next_buffer = p_buffer;
    }
    else
    {
      p_buffer_pool->p_free_head = p_buffer;
    }

    p_buffer_pool->p_free_tail = p_buffer;
  }

  return p_buffer;
}

int32_t ni_logan_buffer_pool_initialize(ni_logan_session_context_t* p_ctx, int32_t number_of_buffers)
{
  int32_t i;

  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  if (p_ctx->buffer_pool != NULL)
  {
    ni_log(NI_LOG_TRACE, "Warn init Buf pool already with size %d\n",
           p_ctx->buffer_pool->number_of_buffers);
    return -1;
  }

  if ((p_ctx->buffer_pool = (ni_logan_queue_buffer_pool_t *) malloc(sizeof(ni_logan_queue_buffer_pool_t))) == NULL)
  {
    ni_log(NI_LOG_ERROR, "Error %d: alloc for pool\n", NI_ERRNO);
    return -1;
  }

  //initialise the struct
  memset(p_ctx->buffer_pool,0,sizeof(ni_logan_queue_buffer_pool_t));
  p_ctx->buffer_pool->number_of_buffers = number_of_buffers;
  //p_buffer_pool->p_free_head = NULL;
  //p_buffer_pool->p_free_tail = NULL;
  //p_buffer_pool->p_used_head = NULL;
  //p_buffer_pool->p_used_tail = NULL;

  for (i = 0; i < number_of_buffers; i++)
  {
    if (NULL == ni_logan_buffer_pool_allocate_buffer(p_ctx->buffer_pool))
    {
      //Release everything we have allocated so far and exit
      ni_logan_buffer_pool_free(p_ctx->buffer_pool);
      return -1;
    }
  }

  return 0;
}

ni_logan_queue_node_t *ni_logan_buffer_pool_expand(ni_logan_queue_buffer_pool_t *pool)
{
  int32_t i;
  for (i = 0; i < 200; i++)
  {
    //TODO: What happens is ni_logan_buffer_pool_allocate_buffer fails and returns a NULL pointer?
    if( NULL == ni_logan_buffer_pool_allocate_buffer(pool) )
    {
      ni_log(NI_LOG_ERROR, "FATAL ERROR: Failed to allocate pool buffer for pool :%p\n", pool);
      return NULL;
    }
  }
  pool->number_of_buffers += 200;
  return pool->p_free_head;
}

ni_logan_queue_node_t *ni_logan_buffer_pool_get_queue_buffer(ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  ni_logan_queue_node_t *buf = NULL;

  if( NULL == p_buffer_pool)
  {
    return NULL;
  }

  buf = p_buffer_pool->p_free_head;

  // find and return a free buffer
  if ( NULL == buf )
  {
    ni_log(NI_LOG_ERROR, "Expanding p_buffer_pool from %d to %d \n",
           p_buffer_pool->number_of_buffers, p_buffer_pool->number_of_buffers + 200);
    buf = ni_logan_buffer_pool_expand(p_buffer_pool);
    if (NULL == buf)
    {
      return NULL; //return null otherwise there will be null derefferencing later
    }
  }

  buf->checkout_timestamp = time(NULL);
  // remove it from free list head; reconnect the linked list, the p_next
  // will become the new head now
  p_buffer_pool->p_free_head = buf->p_next_buffer;

  if ( NULL != buf->p_next_buffer )
  {
    buf->p_next_buffer->p_previous_buffer = NULL;
  }
  else
  {
    p_buffer_pool->p_free_tail = NULL;
  }

  // add it to the used list tail
  buf->p_previous_buffer = p_buffer_pool->p_used_tail;
  buf->p_next_buffer = NULL;

  if ( NULL != p_buffer_pool->p_used_tail )
  {
  p_buffer_pool->p_used_tail->p_next_buffer = buf;
  }
  else
  {
  p_buffer_pool->p_used_head = buf;
  }

  p_buffer_pool->p_used_tail = buf;

  return buf;
}

void ni_logan_buffer_pool_return_buffer(ni_logan_queue_node_t *buf,
                                  ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  if( (!buf) || !(p_buffer_pool) )
  {
    return;
  }

  // remove buf from the used list
  if ( NULL != buf->p_previous_buffer )
  {
    buf->p_previous_buffer->p_next_buffer = buf->p_next_buffer;
  }
  else
  {
    p_buffer_pool->p_used_head = buf->p_next_buffer;
  }

  if ( NULL != buf->p_next_buffer )
  {
    buf->p_next_buffer->p_previous_buffer = buf->p_previous_buffer;
  }
  else
  {
    p_buffer_pool->p_used_tail = buf->p_previous_buffer;
  }

  // put it on the tail of free buffers list
  buf->p_previous_buffer = p_buffer_pool->p_free_tail;
  buf->p_next_buffer = NULL;

  if ( NULL != p_buffer_pool->p_free_tail )
  {
    p_buffer_pool->p_free_tail->p_next_buffer = buf;
  }
  else
  {
    p_buffer_pool->p_free_head = buf;
  }

  p_buffer_pool->p_free_tail = buf;
}

/*!******************************************************************************
*  \brief  Remove a string-pattern from a string in-place.
*
*  \param[in,out] str Null terminated array of characters to operate upon in-place.
*  \param[in] pattern Null terminated array of characters to remove from str.
*                     Supports special characters '#' and '+' for digit matching and
*                     repeated matching respectively. Note, there is no way to \a escape
*                     the special characters.
*                     \b Example:
*                     char a_str[10] = "aaa123qwe";
*                     char b_str[5] = "a+#+";
*                     ni_logan_remove_substring_pattern(a_str, b_str);
*                     printf("%s\n", a_str);
*                     \b Output:
*                     qwe
*
*  \return Output of the pointer to the modified str
*******************************************************************************/
static inline int char_match(char s, char p)
{
  return s == p || p == '#' && isdigit(s);
}

static char *ni_logan_remove_substring_pattern(char *str, char *pattern)
{
  char *start = str;
  char *end = str;
  char *p = pattern;

  if (!str || !pattern || *start == '\0' || *p == '\0' || *p == '+')
  {
    // Invalid string
    return str;
  }

  while (*end != '\0' || *p == '+')
  {
    if (*p == '\0')
    {
      // remove pattern matched substring
      strcpy(start, end);
      end = start;
      p = pattern;
    }
    else if (*p == '+')
    {
      if (char_match(end[-1], p[-1]) && char_match(end[0], p[-1]))
      {
        end++;
      }
      else
      {
        p++;
      }
    }
    else if (char_match(end[0], p[0]))
    {
      end++;
      p++;
    }
    else
    {
      end++;
      start = end;
      p = pattern;
    }
  }

  // remove trailing pattern matched substring
  if (*p == '\0')
  {
    strcpy(start, end);
  }

  return str;
}

#ifdef __ANDROID__
/*!******************************************************************************
*  \brief  use cmd to search nvme block file
*
*  \param[in] p_dev Device name represented as c string. ex: "/dev/nvme0"
*  \param[in] search cmd
*  \param[in] cmd_ret length
*  \param[out] cmd_ret search result for nvme block file
*
*  \return On success returns NI_LOGAN_RETCODE_SUCCESS
*          On failure returns NI_LOGAN_RETCODE_FAILURE
*          On failure returns NI_LOGAN_RETCODE_INVALID_PARAM
*******************************************************************************/
static ni_logan_retcode_t ni_logan_search_file(const char *p_dev, char *cmd, char *cmd_ret, int cmd_ret_len)
{
  FILE *cmd_fp;

  if (access(p_dev, F_OK) == -1)
  {
    return NI_LOGAN_RETCODE_FAILURE;
  }

  // look for child block in sysfs mapping tree
  cmd_fp = popen(cmd, "r");
  if (!cmd_fp)
  {
    return NI_LOGAN_RETCODE_FAILURE;
  }

  if (fgets(cmd_ret, cmd_ret_len, cmd_fp) == NULL)
  {
    pclose(cmd_fp);
    return NI_LOGAN_RETCODE_INVALID_PARAM;
  }

  ni_logan_remove_substring_pattern(cmd_ret, "c#+");

  pclose(cmd_fp);

  return NI_LOGAN_RETCODE_SUCCESS;
}
#endif

/*!******************************************************************************
*  \brief  Find NVMe name space block from device name
*          If none is found, assume nvme multi-pathing is disabled and return /dev/nvmeXn1
*
*  \param[in] p_dev Device name represented as c string. ex: "/dev/nvme0"
*  \param[out] p_out_buf Output buffer to put NVMe name space block. Must be at least length 21
*  \param[in] out_buf_len Length of memory allocated to p_out_buf
*
*  \return On success returns NI_LOGAN_RETCODE_SUCCESS
*          On failure returns NI_LOGAN_RETCODE_FAILURE
*******************************************************************************/
ni_logan_retcode_t ni_logan_find_blk_name(const char *p_dev, char *p_out_buf, int out_buf_len)
{
  FILE *cmd_fp;
  char cmd[128] = {'\0'};
  char cmd_ret[22] = {'\0'};

  if (!p_dev || !p_out_buf)
  {
    return NI_LOGAN_RETCODE_INVALID_PARAM;
  }

#ifdef _WIN32
  ni_log(NI_LOG_TRACE, "In Windows block name equals to device name\n");
  snprintf(p_out_buf, out_buf_len, "%s", p_dev);
  return NI_LOGAN_RETCODE_SUCCESS;
#elif defined(XCODER_LINUX_VIRTIO_DRIVER_ENABLED)
  ni_log(NI_LOG_TRACE, "The device is already considered as a block divice in Linux virtual machine with VirtIO driver.\n");
  snprintf(p_out_buf, out_buf_len, "%s", p_dev);
  return NI_LOGAN_RETCODE_SUCCESS;
#elif defined(__APPLE__)
    snprintf(p_out_buf, out_buf_len, "%s", p_dev);
    return NI_LOGAN_RETCODE_SUCCESS;
#else

#ifdef __ANDROID__
  // assumes no indexing differences between sysfs and udev on Android (ie. no nvme multi-pathing)
  // p_dev[5] is p_dev without '/dev/'
  snprintf(cmd, sizeof(cmd) - 1, "ls /sys/class/nvme/%s/ | grep %s", &p_dev[5], NI_LOGAN_NVME_PREFIX);
#else
  // Note, we are using udevadm through terminal instead of libudev.c to avoid requiring extra kernel dev packges
  snprintf(cmd, sizeof(cmd) - 1, "ls /sys/`udevadm info -q path -n %s` | grep -m 1 -P \"%s(\\d+c)?\\d+n\\d+\"",
           p_dev, NI_LOGAN_NVME_PREFIX);
#endif

  // check p_dev exists in /dev/ folder. If not, return NI_LOGAN_RETCODE_FAILURE
  if (access(p_dev, F_OK) == -1)
  {
    return NI_LOGAN_RETCODE_FAILURE;
  }

  // look for child block in sysfs mapping tree
  cmd_fp = popen(cmd, "r");
  if (!cmd_fp)
  {
    return NI_LOGAN_RETCODE_FAILURE;
  }

  if (fgets(cmd_ret, sizeof(cmd_ret)/sizeof(cmd_ret[0]), cmd_fp) == NULL)
  {
    ni_log(NI_LOG_TRACE, "Failed to find namespaceID. Using guess.\n");
    snprintf(p_out_buf, out_buf_len, "%sn1", p_dev);
  }
  else
  {
    cmd_ret[strcspn(cmd_ret, "\r\n")] = 0;
    ni_logan_remove_substring_pattern(cmd_ret, "c#+");
#ifdef __ANDROID__
    ni_logan_retcode_t ret = NI_LOGAN_RETCODE_SUCCESS;
    snprintf(cmd, sizeof(cmd) - 1, "ls /dev/ | grep %s", cmd_ret); // cmd_ret is block device name
    int cmd_ret_len = sizeof(cmd_ret)/sizeof(cmd_ret[0]);
    ret = ni_logan_search_file(p_dev, cmd, cmd_ret, cmd_ret_len);
    if (ret == NI_LOGAN_RETCODE_SUCCESS)
    {
      char *tmp = NULL;
      if ((tmp = strstr(cmd_ret, "\n")))
      {
        *tmp = '\0';
      }
      snprintf(p_out_buf, out_buf_len, "/dev/%s", cmd_ret);
    }
    else if (ret == NI_LOGAN_RETCODE_INVALID_PARAM)
    {
      snprintf(cmd, sizeof(cmd) - 1, "ls /dev/block/ | grep %s", cmd_ret); // cmd_ret is block device name
      ret = ni_logan_search_file(p_dev, cmd, cmd_ret, cmd_ret_len);
      if (ret == NI_LOGAN_RETCODE_SUCCESS)
      {
        char *tmp = NULL;
        if ((tmp = strstr(cmd_ret, "\n")))
        {
          *tmp = '\0';
        }
        snprintf(p_out_buf, out_buf_len, "/dev/block/%s", cmd_ret);
      }
      else if (ret == NI_LOGAN_RETCODE_INVALID_PARAM)
      {
        ni_log(NI_LOG_ERROR, "Error: %s can not find block device %s\n",
               __FUNCTION__, cmd_ret);
      }
      else
      {
        pclose(cmd_fp);
        return ret;
      }
    }
    else
    {
      pclose(cmd_fp);
      return ret;
    }
#else
    // On systems with virtualized NVMe functions sysfs will include context index in namespaceID.
    // But, udev will not.
    // Thus, it will be necessary to remove the context index from the namespaceID
    // when mapping deviceID to namespaceID through sysfs.
    // ex. change "/dev/nvme0c0n2" to "/dev/nvme0n2"
    snprintf(p_out_buf, out_buf_len, "/dev/%s", cmd_ret);
#endif
  }

  pclose(cmd_fp);

  return NI_LOGAN_RETCODE_SUCCESS;
#endif
}

/*!******************************************************************************
*  \brief  Initialize timestamp handling
*
*  \param
*
*  \return
*******************************************************************************/
ni_logan_retcode_t ni_logan_timestamp_init(ni_logan_session_context_t* p_ctx,
                               ni_logan_timestamp_table_t **pp_table,
                               const char *name)
{
  ni_logan_timestamp_table_t *ptemp;

  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  if (*pp_table != NULL)
  {
    ni_log(NI_LOG_TRACE, "%s: previously allocated, reallocating now\n", __FUNCTION__);
    ni_logan_queue_free(&(*pp_table)->list, p_ctx->buffer_pool);
    free(*pp_table);
  }
  ni_log(NI_LOG_TRACE, "%s: Malloc\n", __FUNCTION__);
  ptemp = (ni_logan_timestamp_table_t *)malloc(sizeof(ni_logan_timestamp_table_t));
  if (!ptemp)
  {
    ni_log(NI_LOG_ERROR, "Error %d: %s\n", NI_ERRNO, __FUNCTION__);
    return NI_LOGAN_RETCODE_ERROR_MEM_ALOC;
  }

  //initialise the struct
  memset(ptemp,0,sizeof(ni_logan_timestamp_table_t));

  ni_logan_queue_init(p_ctx, &ptemp->list, name); //buffer_pool_initialize runs in here

  *pp_table = ptemp;

  ni_log(NI_LOG_TRACE, "%s: success\n", __FUNCTION__);

  return NI_LOGAN_RETCODE_SUCCESS;
}

/*!******************************************************************************
*  \brief  Clean up timestamp handling
*
*  \param
*
*  \return
*******************************************************************************/
ni_logan_retcode_t ni_logan_timestamp_done(ni_logan_timestamp_table_t *p_table,
                               ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  if (!p_table)
  {
    ni_log(NI_LOG_TRACE, "%s: no pts table to free\n", __FUNCTION__);
    return NI_LOGAN_RETCODE_SUCCESS;
  }
  ni_logan_queue_free(&p_table->list, p_buffer_pool);

  free(p_table);
  p_table = NULL;
  ni_log(NI_LOG_TRACE, "%s: success\n", __FUNCTION__);

  return NI_LOGAN_RETCODE_SUCCESS;
}

/*!******************************************************************************
*  \brief  Register timestamp in timestamp/frameoffset table
*
*  \param
*
*  \return
*******************************************************************************/
ni_logan_retcode_t ni_logan_timestamp_register(ni_logan_queue_buffer_pool_t *p_buffer_pool,
                                   ni_logan_timestamp_table_t *p_table,
                                   int64_t timestamp,
                                   uint64_t data_info)
{
  ni_logan_retcode_t err = NI_LOGAN_RETCODE_SUCCESS;

  err = ni_logan_queue_push(p_buffer_pool, &p_table->list, data_info, timestamp);

  if(NI_LOGAN_RETCODE_SUCCESS == err)
  {
    ni_log(NI_LOG_TRACE, "%s: success\n", __FUNCTION__);
  }
  else
  {
    ni_log(NI_LOG_TRACE, "ERROR: %s: FAILED\n", __FUNCTION__);
  }

  return err;
}

/*!******************************************************************************
*  \brief  Retrieve timestamp from table based on frameoffset info
*
*  \param
*
*  \return
*******************************************************************************/
ni_logan_retcode_t ni_logan_timestamp_get(ni_logan_timestamp_table_t *p_table,
                              uint64_t frame_info,
                              int64_t *p_timestamp,
                              int32_t threshold,
                              int32_t print,
                              ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  ni_logan_retcode_t err = NI_LOGAN_RETCODE_SUCCESS;

  ni_log(NI_LOG_TRACE, "%s: getting timestamp with frame_info=%" PRIu64 "\n",
         __FUNCTION__, frame_info);

  err = ni_logan_queue_pop( &p_table->list, frame_info, p_timestamp, threshold, print, p_buffer_pool);
  if(NI_LOGAN_RETCODE_SUCCESS != err)
  {
    ni_log(NI_LOG_TRACE, "%s: error getting timestamp\n", __FUNCTION__);
  }

  ni_log(NI_LOG_TRACE, "%s: timestamp=%" PRId64 ", frame_info=%" PRIu64 ", "
         "err=%d\n", __FUNCTION__, *p_timestamp, frame_info, err);

  return err;
}

ni_logan_retcode_t ni_logan_timestamp_get_with_threshold(ni_logan_timestamp_table_t *p_table,
                                             uint64_t frame_info,
                                             int64_t *p_timestamp,
                                             int32_t threshold,
                                             int32_t print,
                                             ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  return ( ni_logan_queue_pop_threshold(&p_table->list, frame_info, p_timestamp, threshold, print, p_buffer_pool) );
}

void ni_logan_timestamp_scan_cleanup(ni_logan_timestamp_table_t *pts_list,
                               ni_logan_timestamp_table_t *dts_list,
                               ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  ni_logan_queue_t *p_queue = NULL;
  ni_logan_queue_node_t *p = NULL;
  time_t now = time(NULL);

  if (!pts_list || !dts_list)
  {
      return;
  }
  
  // currently, only have dts list.
  // if pts list is added back, this should be modified.
  p_queue = &dts_list->list;
  p = p_queue->p_first;
  
  while (p)
  {
      if (now - p->checkout_timestamp <= 30)
      {
          break;
      }
      if (p_queue->p_first == p_queue->p_last)
      {
          ni_logan_buffer_pool_return_buffer(p_queue->p_first, p_buffer_pool);
          p = p_queue->p_first = p_queue->p_last = NULL;
      } else
      {
          p_queue->p_first = p->p_next;
          p->p_next->p_prev = NULL;
          ni_logan_buffer_pool_return_buffer(p, p_buffer_pool);
  
          p = p_queue->p_first;
      }
      p_queue->count--;
  }
}

/*!******************************************************************************
*  \brief  Retrieve timestamp from table based on frameoffset info
*
*  \param
*
*  \return
*******************************************************************************/
ni_logan_retcode_t ni_logan_timestamp_get_v2(ni_logan_timestamp_table_t *p_table,
                                 uint64_t frame_offset,
                                 int64_t *p_timestamp,
                                 int32_t threshold,
                                 ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  ni_logan_retcode_t err = NI_LOGAN_RETCODE_SUCCESS;

  if( (!p_table) || (!p_timestamp) || (!p_buffer_pool) )
  {
    err = NI_LOGAN_RETCODE_INVALID_PARAM;
    LRETURN;
  }

  ni_log(NI_LOG_TRACE, "%s: getting timestamp with frame_offset=%" PRIu64 "\n",
         __FUNCTION__, frame_offset);

  err = ni_logan_queue_pop(&p_table->list, frame_offset, p_timestamp, threshold, 0, p_buffer_pool);
  if( NI_LOGAN_RETCODE_SUCCESS != err)
  {
    ni_log(NI_LOG_TRACE, "%s: error getting timestamp\n", __FUNCTION__);
  }

  ni_log(NI_LOG_TRACE, "%s: timestamp=%" PRId64 ", frame_offset=%" PRIu64 ", "
         "err=%d\n", __FUNCTION__, *p_timestamp, frame_offset, err);

  END:

  return err;
}

/*!******************************************************************************
*  \brief  Initialize xcoder queue
*
*  \param
*
*  \return
*******************************************************************************/
ni_logan_retcode_t ni_logan_queue_init(ni_logan_session_context_t* p_ctx,
                           ni_logan_queue_t *p_queue,
                           const char *name)
{
  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  if (!p_queue || !name)
  {
    return NI_LOGAN_RETCODE_INVALID_PARAM;
  }
  p_queue->name[0] = '\0';
  strncat(p_queue->name, name, strlen(name));
  ni_logan_buffer_pool_initialize(p_ctx, LOGAN_BUFFER_POOL_SZ_PER_CONTEXT);

  p_queue->p_first = NULL;
  p_queue->p_last = NULL;
  p_queue->count = 0;

  ni_log(NI_LOG_TRACE, "%s: exit\n", __FUNCTION__);

  return NI_LOGAN_RETCODE_SUCCESS;
}

/*!******************************************************************************
*  \brief  Push into xcoder queue
*
*  \param
*
*  \return
*******************************************************************************/
ni_logan_retcode_t ni_logan_queue_push(ni_logan_queue_buffer_pool_t *p_buffer_pool,
                           ni_logan_queue_t *p_queue,
                           uint64_t frame_info,
                           int64_t timestamp)
{
  ni_logan_retcode_t err = NI_LOGAN_RETCODE_SUCCESS;
  ni_logan_queue_node_t *temp = NULL;

  if ( !p_queue )
  {
    ni_log(NI_LOG_TRACE, "%s: error, null pointer parameters passed\n",
           __FUNCTION__);
    err = NI_LOGAN_RETCODE_INVALID_PARAM;
    LRETURN;
  }

  temp = ni_logan_buffer_pool_get_queue_buffer(p_buffer_pool);

  if (!temp)
  {
    ni_log(NI_LOG_TRACE, "%s: error, cannot allocate memory\n", __FUNCTION__);
    err = NI_LOGAN_RETCODE_ERROR_MEM_ALOC;
    LRETURN;
  }

  temp->timestamp = timestamp;
  temp->frame_info = frame_info;

  temp->p_next = NULL;

  if (!p_queue->p_first)
  {
    p_queue->p_first = p_queue->p_last = temp;
    p_queue->p_first->p_prev = NULL;
  }
  else
  {
    p_queue->p_last->p_next = temp;
    temp->p_prev = p_queue->p_last;
    p_queue->p_last = temp;
  }

  p_queue->count++;

  if (p_queue->count > LOGAN_XCODER_MAX_NUM_QUEUE_ENTRIES)
  {
    ni_log(NI_LOG_DEBUG, "%s: queue overflow, remove oldest entry, count=%d\n",
           __FUNCTION__, p_queue->count);
    // ni_logan_assert(0);
    //Remove oldest one
    temp = p_queue->p_first->p_next;
    // free(p_queue->p_first);
    ni_logan_buffer_pool_return_buffer(p_queue->p_first, p_buffer_pool);
    p_queue->p_first = temp;
    p_queue->p_first->p_prev = NULL;
    p_queue->count--;
  }

  END:

  return err;
}

/*!******************************************************************************
*  \brief  Pop from the xcoder queue
*
*  \param
*
*  \return
*******************************************************************************/
ni_logan_retcode_t ni_logan_queue_pop(ni_logan_queue_t *p_queue,
                          uint64_t frame_info,
                          int64_t *p_timestamp,
                          int32_t threshold,
                          int32_t print,
                          ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  ni_logan_queue_node_t *temp;
  ni_logan_queue_node_t *temp_prev = NULL;
  int32_t found = 0;
  int32_t count = 0;
  ni_logan_retcode_t retval = NI_LOGAN_RETCODE_SUCCESS;

  if (!p_queue || !p_timestamp)
  {
    ni_log(NI_LOG_TRACE, "%s: error, null pointer parameters passed\n",
           __FUNCTION__);
    retval = NI_LOGAN_RETCODE_INVALID_PARAM;
    LRETURN;
  }

  if ( NULL == p_queue->p_first )
  {
    ni_log(NI_LOG_TRACE, "%s: queue is empty...\n", __FUNCTION__);
    retval =  NI_LOGAN_RETCODE_FAILURE;
    LRETURN;
  }

  if (p_queue->p_first == p_queue->p_last)
  {
    /*! If only one entry, retrieve timestamp without checking */
    *p_timestamp = p_queue->p_first->timestamp;
    // free(p_queue->p_first);
    ni_logan_buffer_pool_return_buffer(p_queue->p_first, p_buffer_pool);

    p_queue->p_first = NULL;
    p_queue->p_last = NULL;
    p_queue->count--;
    ni_logan_assert(p_queue->count == 0);
    found = 1;
  }
  else
  {
    temp = p_queue->p_first;
    while (temp && !found)
    {
      if (frame_info < temp->frame_info)
      {
        if (!temp->p_prev)
        {
          ni_log(NI_LOG_TRACE, "First in ts list, return it\n");
          *p_timestamp = temp->timestamp;

          p_queue->p_first = temp->p_next;
          temp->p_next->p_prev = NULL;

          ni_logan_buffer_pool_return_buffer(temp, p_buffer_pool);
          p_queue->count--;
          found = 1;
          break;
        }

        // retrieve from p_prev and delete p_prev !
        *p_timestamp = temp->p_prev->timestamp;
        temp = temp->p_prev;
        temp_prev = temp->p_prev;

        if (!temp_prev)
        {
          p_queue->p_first = temp->p_next;
          temp->p_next->p_prev = NULL;
        }
        else
        {
          temp_prev->p_next = temp->p_next;
          if (temp->p_next)
          {
            temp->p_next->p_prev = temp_prev;
          }
          else
          {
            p_queue->p_last = temp_prev;
          }
        }
        //free(temp);
        ni_logan_buffer_pool_return_buffer(temp, p_buffer_pool);
        p_queue->count--;
        found = 1;
        break;
      }
      temp_prev = temp;
      temp = temp->p_next;
      count++;
    }
  }

  if (print)
  {
    ni_log(NI_LOG_TRACE, "%s %s %d iterations ..\n",
           __FUNCTION__, p_queue->name, count);
  }
  if (p_queue->count < 0)
    {
    ni_log(NI_LOG_TRACE, "%s: error p_queue->count=%d\n",
           __FUNCTION__, p_queue->count);
    ni_logan_assert(0);
  }

  if (!found)
  {
    retval = NI_LOGAN_RETCODE_FAILURE;
  }

  END:

  return retval;
}

ni_logan_retcode_t ni_logan_queue_pop_threshold(ni_logan_queue_t *p_queue,
                                    uint64_t frame_info,
                                    int64_t *p_timestamp,
                                    int32_t threshold,
                                    int32_t print,
                                    ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  ni_logan_queue_node_t *temp;
  ni_logan_queue_node_t *temp_prev = NULL;
  int32_t found = 0;
  int32_t count = 0;
  ni_logan_retcode_t retval = NI_LOGAN_RETCODE_SUCCESS;

  if ( (!p_queue) || (!p_timestamp) )
  {
    ni_log(NI_LOG_TRACE, "%s: error, null pointer parameters passed\n",
           __FUNCTION__);
    retval = NI_LOGAN_RETCODE_INVALID_PARAM;
    LRETURN;
  }

  if (p_queue->p_first == NULL)
  {
    ni_log(NI_LOG_TRACE, "%s: queue is empty...\n", __FUNCTION__);
    retval = NI_LOGAN_RETCODE_FAILURE;
    LRETURN;
  }

  if (p_queue->p_first == p_queue->p_last)
  {
    /*! If only one entry, retrieve timestamp without checking */
    *p_timestamp = p_queue->p_first->timestamp;
    // free(p_queue->p_first);
    ni_logan_buffer_pool_return_buffer(p_queue->p_first, p_buffer_pool);
    p_queue->p_first = NULL;
    p_queue->p_last = NULL;
    p_queue->count--;
    found = 1;
  }
  else
  {
    temp = p_queue->p_first;
    while (temp && !found)
    {
      if (llabs(frame_info - temp->frame_info) <= threshold)
      {
        *p_timestamp = temp->timestamp;
        if (!temp_prev)
        {
          p_queue->p_first = temp->p_next;
          temp->p_next->p_prev = NULL;
        }
        else
        {
          temp_prev->p_next = temp->p_next;
          if (temp->p_next)
          {
            temp->p_next->p_prev = temp_prev;
          }
          else
          {
            p_queue->p_last = temp_prev;
          }
        }
        // free(temp);
        ni_logan_buffer_pool_return_buffer(temp, p_buffer_pool);
        p_queue->count--;
        found = 1;
        break;
      }
      temp_prev = temp;
      temp = temp->p_next;
      count++;
    }
  }
  if (print)
  {
    ni_log(NI_LOG_TRACE, "%s %s %d iterations ..\n",
           __FUNCTION__, p_queue->name, count);
  }

  if (p_queue->count < 0)
  {
    ni_log(NI_LOG_TRACE, "%s: error p_queue->count=%d\n",
           __FUNCTION__, p_queue->count);
  }

  if (!found)
  {
    retval =  NI_LOGAN_RETCODE_FAILURE;
  }

  END:

  return retval;
}

/*!******************************************************************************
*  \brief  Free xcoder queue
*
*  \param
*
*  \return
*******************************************************************************/
ni_logan_retcode_t ni_logan_queue_free(ni_logan_queue_t *p_queue, ni_logan_queue_buffer_pool_t *p_buffer_pool)
{
  ni_logan_queue_node_t *temp = NULL;
  ni_logan_queue_node_t *temp_next;
  int32_t left = 0;

  if (!p_queue)
  {
    return NI_LOGAN_RETCODE_SUCCESS;
  }

  ni_log(NI_LOG_TRACE, "Entries before clean up: \n");
  ni_logan_queue_print(p_queue);

  temp = p_queue->p_first;
  while (temp)
  {
    temp_next = temp->p_next;
    //free(temp);
    ni_logan_buffer_pool_return_buffer(temp, p_buffer_pool);
    temp = temp_next;
    left++;
  }
  ni_log(NI_LOG_TRACE, "Entries cleaned up at ni_logan_queue_free: %d, count: %d\n",
         left, p_queue->count);

  //ni_logan_queue_print(p_queue);

  p_queue->count = 0;

  return NI_LOGAN_RETCODE_SUCCESS;
}

/*!******************************************************************************
*  \brief  Print xcoder queue info
*
*  \param
*
*  \return
*******************************************************************************/
ni_logan_retcode_t ni_logan_queue_print(ni_logan_queue_t *p_queue)
{
  ni_logan_queue_node_t *temp = NULL;
  struct tm* ltime = NULL;
  char buff[20] = { 0 };

  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  if (!p_queue)
  {
    return NI_LOGAN_RETCODE_SUCCESS;
  }

  ni_log(NI_LOG_TRACE, "Queue [%s] Count: %d\n", p_queue->name, p_queue->count);

  ni_log(NI_LOG_TRACE, "\nForward:\n");

  temp = p_queue->p_first;

  ni_log(NI_LOG_TRACE, "%s: p_first=%p, p_last=%p, count=%d, temp=%p\n",
         __FUNCTION__, p_queue->p_first, p_queue->p_last, p_queue->count, temp);

  while (temp)
  {
    ltime = localtime(&temp->checkout_timestamp);
    if (ltime)
    {
      strftime(buff, 20, "%Y-%m-%d %H:%M:%S", ltime);
      ni_log(NI_LOG_TRACE, " %s [%" PRIu64 ", %" PRIu64 "]", buff, temp->timestamp, temp->frame_info);
    }
    temp = temp->p_next;
  }

  ni_log(NI_LOG_TRACE, "\nBackward:");

  temp = p_queue->p_last;
  while (temp)
  {
    ni_log(NI_LOG_TRACE, " [%" PRIu64 ", %" PRIu64 "]\n", temp->timestamp, temp->frame_info);
    temp = temp->p_prev;
  }
  ni_log(NI_LOG_TRACE, "\n");

  return NI_LOGAN_RETCODE_SUCCESS;
}

/*!******************************************************************************
 *  \brief Initialize a fifo buffer
 *
 *  \param[in] number_of_buffers  Total number of buffers in this fifo
 *  \param[in] size               The size of every buffer
 *
 *  \return On success return fifo pointer, On failure return NULL
 *******************************************************************************/
ni_logan_fifo_buffer_t * ni_logan_fifo_initialize(uint32_t number_of_buffers,
                                                  uint32_t size)
{
  ni_logan_fifo_buffer_t *p_fifo = NULL;

  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  if ((p_fifo = (ni_logan_fifo_buffer_t *) malloc(sizeof(ni_logan_fifo_buffer_t))) == NULL)
  {
    ni_log(NI_LOG_ERROR, "Error %d: %s alloc for fifo\n", NI_ERRNO, __FUNCTION__);
    return NULL;
  }

  //initialise the struct
  memset(p_fifo,0,sizeof(ni_logan_fifo_buffer_t));
  ni_logan_pthread_mutex_init(&(p_fifo->mutex));
  p_fifo->number_of_buffers = number_of_buffers;
  p_fifo->buffer_size = size;

  // alloc fifo buffer
  p_fifo->buffer = malloc(number_of_buffers * size);
  if (p_fifo->buffer == NULL)
  {
    free(p_fifo);
    p_fifo = NULL;
    ni_log(NI_LOG_ERROR, "Error %d: %s alloc for fifo buffer\n", NI_ERRNO, __FUNCTION__);
    return NULL;
  }
  memset(p_fifo->buffer,0,(number_of_buffers * size));

  return p_fifo;
}

/*!******************************************************************************
 *  \brief Free a fifo
 *
 *  \param[in] p_fifo  The pointer of ni_logan_fifo_buffer_t
 *
 *  \return
 *******************************************************************************/
void ni_logan_fifo_free(ni_logan_fifo_buffer_t * p_fifo)
{
  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  if (!p_fifo)
  {
    ni_log(NI_LOG_TRACE, "%s: null fifo, return\n", __FUNCTION__);
    return;
  }
  ni_logan_pthread_mutex_destroy(&(p_fifo->mutex));

  free(p_fifo->buffer);
  p_fifo->buffer = NULL;

  free(p_fifo);

  return;
}

/*!******************************************************************************
 *  \brief Check if a fifo is full
 *
 *  \param[in] p_fifo  The pointer of ni_logan_fifo_buffer_t
 *
 *  \return 1 is full; 0 is not full
 *******************************************************************************/
int ni_logan_fifo_is_full(ni_logan_fifo_buffer_t * p_fifo)
{
  return (p_fifo->number_of_buffers_used == p_fifo->number_of_buffers);
}

/*!******************************************************************************
 *  \brief Check if a fifo is empty
 *
 *  \param[in] p_fifo  The pointer of ni_logan_fifo_buffer_t
 *
 *  \return 1 is empty; 0 is not empty
 *******************************************************************************/
int ni_logan_fifo_is_empty(ni_logan_fifo_buffer_t * p_fifo)
{
  return (p_fifo->number_of_buffers_used == 0);
}

/*!******************************************************************************
 *  \brief Get free buffer to write in the fifo
 *
 *  \param[in] p_fifo  The pointer of ni_logan_fifo_buffer_t
 *
 *  \return On success return the pointer of free buffer,
 *          On failure return NULL
 *******************************************************************************/
uint8_t *ni_logan_fifo_generic_write(ni_logan_fifo_buffer_t * p_fifo)
{
  uint8_t *p_buffer;
  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  if (NULL == p_fifo || ni_logan_fifo_is_full(p_fifo))
  {
    ni_log(NI_LOG_TRACE, "Warning pointer is null or fifo is full, %p\n", p_fifo);
    return NULL;
  }

  ni_logan_pthread_mutex_lock(&(p_fifo->mutex));

  p_buffer = p_fifo->buffer + p_fifo->wndx * p_fifo->buffer_size;
  p_fifo->wndx = (p_fifo->wndx + 1) % p_fifo->number_of_buffers;
  p_fifo->number_of_buffers_used++;

  ni_logan_pthread_mutex_unlock(&(p_fifo->mutex));
  return p_buffer;
}

/*!******************************************************************************
 *  \brief Get first filled buffer to read in the fifo
 *
 *  \param[in] p_fifo  The pointer of ni_logan_fifo_buffer_t
 *
 *  \return On success return the pointer of filled buffer in fifo,
 *          On failure return NULL
 *******************************************************************************/
uint8_t *ni_logan_fifo_generic_read(ni_logan_fifo_buffer_t * p_fifo)
{
  uint8_t *p_buffer;
  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  if (NULL == p_fifo || ni_logan_fifo_is_empty(p_fifo))
  {
    ni_log(NI_LOG_TRACE, "Warning pointer is null or fifo is empty, %p\n", p_fifo);
    return NULL;
  }

  ni_logan_pthread_mutex_lock(&(p_fifo->mutex));
  p_buffer = p_fifo->buffer + p_fifo->rndx * p_fifo->buffer_size;
  p_fifo->rndx = (p_fifo->rndx + 1) % p_fifo->number_of_buffers;
  p_fifo->number_of_buffers_used--;

  ni_logan_pthread_mutex_unlock(&(p_fifo->mutex));
  return p_buffer;
}

/*!******************************************************************************
 *  \brief Push back the last read buffer to the fifo
 *
 *  \param[in] p_fifo  The pointer of ni_logan_fifo_buffer_t
 *
 *  \return On success return 0,
 *          On failure return -1
 *******************************************************************************/
int ni_logan_fifo_return_read(ni_logan_fifo_buffer_t * p_fifo)
{
  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  if (NULL == p_fifo || ni_logan_fifo_is_full(p_fifo))
  {
    ni_log(NI_LOG_TRACE, "Warning pointer is null or fifo is full, %p\n", p_fifo);
    return -1;
  }

  ni_logan_pthread_mutex_lock(&(p_fifo->mutex));

  p_fifo->rndx = p_fifo->rndx > 0 ? (p_fifo->rndx - 1) : (p_fifo->number_of_buffers - 1);
  p_fifo->number_of_buffers_used++;

  ni_logan_pthread_mutex_unlock(&(p_fifo->mutex));
  return 0;
}

/*!******************************************************************************
*  \brief  Convert string to boolean
*
*  \param
*
*  \return
*******************************************************************************/
int32_t ni_logan_atobool(const char *p_str, bool *b_error)
{
  if (!strcmp(p_str, "1") ||
      !strcmp(p_str, "true") ||
      !strcmp(p_str, "yes"))
  {
    return 1;
  }

  if (!strcmp(p_str, "0") ||
      !strcmp(p_str, "false") ||
      !strcmp(p_str, "no"))
  {
    return 0;
  }

  *b_error = true;
  return 0;
}

/*!******************************************************************************
*  \brief  Convert string to integer
*
*  \param
*
*  \return
*******************************************************************************/
int32_t ni_logan_atoi(const char *p_str, bool *b_error)
{
  char *end;
  int32_t v = strtol(p_str, &end, 0);

  if (end == p_str || *end != '\0')
  {
    *b_error = true;
  }

  return v;
}

/*!******************************************************************************
*  \brief  Convert string to floating
*
*  \param
*
*  \return
*******************************************************************************/
double ni_logan_atof(const char *p_str, bool *b_error)
{
  char *end;
  double v = strtod(p_str, &end);

  if (end == p_str || *end != '\0')
  {
    *b_error = true;
  }

  return v;
}

/*!******************************************************************************
*  \brief  Parse name
*
*  \param
*
*  \return
*******************************************************************************/
int32_t ni_logan_parse_name(const char *arg, const char *const *names, bool *b_error)
{
  int32_t i;
  for (i = 0; names[i]; i++)
  {
    if (!strcmp(arg, names[i]))
    {
      return i;
    }
  }

  return ni_logan_atoi(arg, b_error);
}

/*!*****************************************************************************
*  \brief   threadpool control
*
*  \param[in]   params
*
*  \return  NULL
*
*
******************************************************************************/
void *thread_routine(void *arg)
{
  // ni_logan_pthread create start
  threadpool_t *pool = (threadpool_t *)arg;
  // Loop the pool fetch the thread to run
  ni_log(NI_LOG_TRACE, "thread is starting\n");
  for (; ;)
  {
    //set mutex
    ni_logan_pthread_mutex_lock(&pool->pmutex);
    pool->idle++;
    //pool is empty and pool is not quit then wait add timeout，timeout=1
    while (pool->first == NULL && !pool->quit)
    {
      ni_log(NI_LOG_TRACE, "thread_routine is waiting\n");
      ni_logan_pthread_cond_wait(&pool->pcond, &pool->pmutex);
    }

    // the pool is not empty
    // idle--，start run the task
    pool->idle--;
    if (pool->first != NULL)
    {
      ni_task_t *t = pool->first;
      pool->first = t->next;
      ni_logan_pthread_mutex_unlock(&pool->pmutex);
      t->run(t->arg);
      free(t);
      ni_logan_pthread_mutex_lock(&pool->pmutex);
    }
    // pool is quit and the pool is empty,  break
    if (pool->quit && pool->first == NULL)
    {
      pool->counter--;
      if (pool->counter == 0)
      {
        ni_logan_pthread_cond_signal(&pool->pcond);
      }
      ni_logan_pthread_mutex_unlock(&pool->pmutex);
      break;
    }
    ni_logan_pthread_mutex_unlock(&pool->pmutex);
  }

  ni_log(NI_LOG_TRACE, "thread_routine is exiting\n");
  return NULL;
}

/*!*****************************************************************************
*  \brief   Init the threadpool
*
*  \param[in]   pool  threadpool address
*
*  \return      NULL
*
*
******************************************************************************/
void threadpool_init(threadpool_t *pool)
{
  ni_logan_pthread_mutex_init(&pool->pmutex);
  ni_logan_pthread_cond_init(&pool->pcond, NULL);
  pool->first = NULL;
  pool->last = NULL;
  pool->counter = 0;
  pool->idle = 0;
  pool->max_threads = MAX_THREADS;
  pool->quit = 0;
}

/*!*****************************************************************************
*  \brief   add task to threadpool using newThread control it
*
*  \param[in]   pool        threadpool address
*  \param[in]   run         run function
*  \param[in]   arg         run function params
*  \param[in]   newThread   1: create a new thread.
*                           0: do not create a new thread.
*
*  \return      0     success
*              <0     failed
*
*
******************************************************************************/
int threadpool_auto_add_task_thread(threadpool_t *pool,
                                    void *(*run)(void *arg),
                                    void *arg,
                                    int newThread)
{
  int ret = 0;
  // create new task
  ni_task_t *newtask = (ni_task_t *)malloc(sizeof(ni_task_t));
  if (!newtask)
  {
    ni_log(NI_LOG_ERROR, "ERROR %d: %s Failed to allocate memory\n",
           NI_ERRNO, __FUNCTION__);
    return -1;
  }
  newtask->run = run;
  newtask->arg = arg;
  newtask->next = NULL;

  // add mutex
  ni_logan_pthread_mutex_lock(&pool->pmutex);
  // add newtask to the pool
  if (pool->first == NULL)
  {
    pool->first = newtask;
  }
  else
  {
    pool->last->next = newtask;
  }
  pool->last = newtask;

  // If their are idle tasks, wake up the tasks
  ni_log(NI_LOG_TRACE, "%s pool->idle %d\n", __FUNCTION__, pool->idle);
  if (pool->idle > 0)
  {
    ni_logan_pthread_cond_signal(&pool->pcond);
  }
  // no idle tasks，create a new thread，run the thread_routine
  else
  {
    if (pool->counter < pool->max_threads)
    {
      if (newThread == 1)
      {
        ni_pthread_t tid;
        ni_logan_pthread_create(&tid, NULL, thread_routine, pool);
        if (ret)
        {
          ni_log(NI_LOG_ERROR, "ERROR %d: %s ni_logan_pthread_create failed : %d\n",
                 NI_ERRNO, __FUNCTION__, ret);
          return -1;
        }
        pool->counter++;
      }
    }
    else
    {
      while (pool->idle == 0)
      {
        ni_logan_usleep(1000);
      }
      ni_logan_pthread_cond_signal(&pool->pcond);
    }
  }
  ni_log(NI_LOG_TRACE, "%s pool->counter %d\n", __FUNCTION__, pool->counter);
  ni_logan_pthread_mutex_unlock(&pool->pmutex);
  return ret;
}


/*!*****************************************************************************
*  \brief   add task to threadpool
*
*  \param[in]   pool  threadpool address
*  \param[in]   run   run function
*  \param[in]   arg   run function params
*
*  \return      0     success
*              <0     failed
*
*
******************************************************************************/
int threadpool_add_task(threadpool_t *pool, void *(*run)(void *arg), void *arg)
{
  int ret = 0;
  // create new task
  ni_task_t *newtask = (ni_task_t *)malloc(sizeof(ni_task_t));
  if (!newtask)
  {
    ni_log(NI_LOG_ERROR, "ERROR %d: %s Failed to allocate memory\n",
           NI_ERRNO, __FUNCTION__);
    return -1;
  }
  newtask->run = run;
  newtask->arg = arg;
  newtask->next = NULL;

  // add mutex
  ni_logan_pthread_mutex_lock(&pool->pmutex);
  // add newtask to the pool
  if (pool->first == NULL)
  {
    pool->first = newtask;
  }
  else
  {
    pool->last->next = newtask;
  }
  pool->last = newtask;

  // If their are idle tasks, wake up the tasks
  ni_log(NI_LOG_TRACE, "%s pool->idle %d\n", __FUNCTION__, pool->idle);
  if (pool->idle > 0)
  {
    ni_logan_pthread_cond_signal(&pool->pcond);
  }
  // no idle tasks，create a new thread，run the thread_routine
  else if (pool->counter < pool->max_threads)
  {
    ni_pthread_t tid;
    ret = ni_logan_pthread_create(&tid, NULL, thread_routine, pool);
    if (ret)
    {
      ni_log(NI_LOG_ERROR, "ERROR %d: %s ni_logan_pthread_create failed : %d\n",
             NI_ERRNO, __FUNCTION__, ret);
      return -1;
    }
    pool->counter++;
  }
  ni_log(NI_LOG_TRACE, "%s pool->counter %d\n", __FUNCTION__, pool->counter);
  ni_logan_pthread_mutex_unlock(&pool->pmutex);
  return ret;
}


/*!*****************************************************************************
*  \brief   destroy threadpool
*
*  \param[in]   pool  threadpool address
*
*  \return  NULL
*
*
******************************************************************************/
void threadpool_destroy(threadpool_t *pool)
{
  ni_log(NI_LOG_TRACE, "destroy start!\n");
  if (pool->quit)
  {
    return;
  }
  // add mutex
  ni_logan_pthread_mutex_lock(&pool->pmutex);
  pool->quit = 1;

  if (pool->counter > 0)
  {
    // if has idle task, wake up them
    if (pool->idle > 0)
    {
      ni_log(NI_LOG_TRACE, "destroy broadcast!\n");
      // wake up all the idle tasks
      ni_logan_pthread_cond_broadcast(&pool->pcond);
    }
    // if the process is running, wait it finish
    while (pool->counter)
    {
      ni_logan_pthread_cond_wait(&pool->pcond, &pool->pmutex);
    }
  }
  ni_logan_pthread_mutex_unlock(&pool->pmutex);
  ni_logan_pthread_mutex_destroy(&pool->pmutex);
  ni_logan_pthread_cond_destroy(&pool->pcond);
}

// Netint HW YUV420p data layout related utility functions

/*!*****************************************************************************
*  \brief  Get dimension information of Netint HW YUV420p frame to be sent
*          to encoder for encoding. Caller usually retrieves this info and
*          uses it in the call to ni_logan_encoder_frame_buffer_alloc for buffer
*          allocation.
*
*  \param[in]  width   source YUV frame width
*  \param[in]  height  source YUV frame height
*  \param[in]  bit_depth_factor  1 for 8 bit, 2 for 10 bit
*  \param[in]  is_h264  non-0 for H.264 codec, 0 otherwise (H.265)
*  \param[out] plane_stride  size (in bytes) of each plane width
*  \param[out] plane_height  size of each plane height
*
*  \return Y/Cb/Cr stride and height info
*
******************************************************************************/
void ni_logan_get_hw_yuv420p_dim(int width, int height, int bit_depth_factor,
                           int is_h264,
                           int plane_stride[NI_LOGAN_MAX_NUM_DATA_POINTERS],
                           int plane_height[NI_LOGAN_MAX_NUM_DATA_POINTERS])
{
  // width (in pixels) is multiples of 32
  plane_stride[0] = ((width + 31) / 32) * 32;
  if (plane_stride[0] < NI_LOGAN_MIN_WIDTH)
  {
    plane_stride[0] = NI_LOGAN_MIN_WIDTH;
  }
  plane_stride[0] *= bit_depth_factor;
  plane_stride[1] = plane_stride[2] = plane_stride[0] / 2;

  // height (in lines) is multiples of 8 for H.265, 16 for H.264
  plane_height[0] = ((height + 7) / 8) * 8;
  if (is_h264)
  {
    plane_height[0] = ((height + 15) / 16) * 16;
  }
  if (plane_height[0] < NI_LOGAN_MIN_HEIGHT)
  {
    plane_height[0] = NI_LOGAN_MIN_HEIGHT;
  }
  plane_height[1] = plane_height[2] = plane_height[0] / 2;
}

/*!*****************************************************************************
*  \brief  Copy YUV data to Netint HW YUV420p frame layout to be sent
*          to encoder for encoding. Data buffer (dst) is usually allocated by
*          ni_logan_encoder_frame_buffer_alloc.
*
*  \param[out] p_dst  pointers of Y/Cb/Cr to which data is copied
*  \param[in]  p_src  pointers of Y/Cb/Cr from which data is copied
*  \param[in]  width  source YUV frame width
*  \param[in]  height source YUV frame height
*  \param[in]  bit_depth_factor  1 for 8 bit, 2 for 10 bit
*  \param[in]  dst_stride  size (in bytes) of each plane width in destination
*  \param[in]  dst_height  size of each plane height in destination
*  \param[in]  src_stride  size (in bytes) of each plane width in source
*  \param[in]  src_height  size of each plane height in source
*
*  \return Y/Cb/Cr data
*
******************************************************************************/
void ni_logan_copy_hw_yuv420p(uint8_t *p_dst[NI_LOGAN_MAX_NUM_DATA_POINTERS],
                        uint8_t *p_src[NI_LOGAN_MAX_NUM_DATA_POINTERS],
                        int frame_width, int frame_height, int bit_depth_factor,
                        int dst_stride[NI_LOGAN_MAX_NUM_DATA_POINTERS],
                        int dst_height[NI_LOGAN_MAX_NUM_DATA_POINTERS],
                        int src_stride[NI_LOGAN_MAX_NUM_DATA_POINTERS],
                        int src_height[NI_LOGAN_MAX_NUM_DATA_POINTERS])
{
  // return to avoid self copy
  if (p_dst[0] == p_src[0] && p_dst[1] == p_src[1] && p_dst[2] == p_src[2])
  {
    ni_log(NI_LOG_TRACE, "%s: src and dst identical, return\n", __FUNCTION__);
    return;
  }

  ni_log(NI_LOG_TRACE, "%s dst_stride %d/%d/%d src_stride %d/%d/%d dst_height "
         "%d/%d/%d src_height %d/%d/%d\n", __FUNCTION__,
         dst_stride[0], dst_stride[1], dst_stride[2],
         src_stride[0], src_stride[1], src_stride[2],
         dst_height[0], dst_height[1], dst_height[2],
         src_height[0], src_height[1], src_height[2]);

  // width padding length in pixels and bytes, if needed
  int pad_len = 0, pad_len_bytes;
  int i;

  for (i = 0; i < NI_LOGAN_MAX_NUM_DATA_POINTERS; i++)
  {
    int height = src_height[i] < dst_height[i] ? src_height[i] : dst_height[i];
    uint8_t *dst = p_dst[i];
    const uint8_t *src = (const uint8_t *)p_src[i];

    if (0 == i) // Y
    {
      pad_len = dst_stride[i] / bit_depth_factor - frame_width;
    }
    else if (1 == i)
    {
      // U/V share the same padding length; Y padding could be odd value
      // so make it even
      pad_len = ((pad_len + 1) / 2) * 2;
    }

    pad_len_bytes = (i == 0 ? pad_len * bit_depth_factor :
    pad_len * bit_depth_factor / 2);
    ni_log(NI_LOG_TRACE, "%s plane %d stride padding: %d pixel (%d bytes), "
           "copy height: %d.\n", __FUNCTION__, i, pad_len, pad_len_bytes, height);

    if(pad_len || src_stride[i] != dst_stride[i])
    {
      for (; height > 0; height--)
      {
        memcpy(dst, src, (src_stride[i] < dst_stride[i] ?
        src_stride[i] : dst_stride[i]));
        dst += dst_stride[i];

        // dst is now at the line end
        // repeat last pixel
        if (2 == bit_depth_factor)
        {
          // for 10 bit it's 2 bytes
          int j;
          uint8_t *tmp_dst = dst - pad_len_bytes;
          for (j = 0; j < pad_len_bytes / 2; j++)
          {
            memcpy(tmp_dst, dst - pad_len_bytes - 2, 2);
            tmp_dst += 2;
          }
        }
        else
        {
          memset(dst - pad_len_bytes, *(dst - pad_len_bytes - 1),
          pad_len_bytes);
        }
        src += src_stride[i];
      }
    }
    else
    {
      int copy_stride = dst_stride[i];
      memcpy(dst, src, copy_stride * height);
      dst += copy_stride * height;
    }

    // height padding/cropping if needed
    int padding_height = dst_height[i] - src_height[i];
    if (padding_height > 0)
    {
      ni_log(NI_LOG_TRACE, "%s plane %d padding height: %d\n",
             __FUNCTION__, i, padding_height);
      src = dst - dst_stride[i];
      for (; padding_height > 0; padding_height--)
      {
        memcpy(dst, src, dst_stride[i]);
        dst += dst_stride[i];
      }
    }
  }
}

// NAL operations

/*!*****************************************************************************
*  \brief  Insert emulation prevention byte(s) as needed into the data buffer
*
*  \param  buf   data buffer to be worked on - new byte(s) will be inserted
*          size  number of bytes starting from buf to check
*
*  \return the number of emulation prevention bytes inserted into buf, 0 if
*          none.
*
*  Note: caller *MUST* ensure for newly inserted bytes, buf has enough free
*        space starting from buf + size
******************************************************************************/
int ni_logan_insert_emulation_prevent_bytes(uint8_t *buf, int size)
{
  int insert_bytes = 0;
  uint8_t *buf_curr = buf;
  uint8_t *buf_end = buf + size - 1;
  int zeros = 0, insert_ep3_byte = 0;

  ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

  for (; buf_curr <= buf_end; buf_curr++)
  {
    if (zeros == 2)
    {
      insert_ep3_byte = (*buf_curr <= 3);
      if (insert_ep3_byte)
      {
        // move bytes from curr to end 1 position to make space for ep3
        memmove(buf_curr + 1, buf_curr, buf_end - buf_curr + 1);
        *buf_curr = 0x3;
        buf_curr++;
        buf_end++;
        insert_bytes++;
      }

      zeros = 0;
    }

    if (! *buf_curr)
    {
      zeros++;
    }
    else
    {
      zeros = 0;
    }
  }

  ni_log(NI_LOG_TRACE, "%s: %d, exit\n", __FUNCTION__, insert_bytes);
  return insert_bytes;
}

/*!*****************************************************************************
 *  \brief  Remove emulation prevention byte(s) as needed from the data buffer
 *
 *  \param  buf   data buffer to be worked on - emu prevent byte(s) will be
 *                removed from.
 *          size  number of bytes starting from buf to check
 *
 *  \return the number of emulation prevention bytes removed from buf, 0 if
 *          none.
 *
 *  Note: buf will be modified if emu prevent byte(s) found and removed.
 ******************************************************************************/
LIB_API int ni_logan_remove_emulation_prevent_bytes(uint8_t *buf, int size)
{
    int remove_bytes = 0;
    uint8_t *buf_curr = buf;
    uint8_t *buf_end = buf + size - 1;
    int zeros = 0, remove_ep3_byte;

    ni_log(NI_LOG_TRACE, "%s: enter\n", __func__);

    for (; buf_curr < buf_end; buf_curr++)
    {
        if (zeros == 2)
        {
            remove_ep3_byte = (*buf_curr == 0x03 && *(buf_curr + 1) <= 3);
            if (remove_ep3_byte)
            {
                // move bytes after curr to end, 1 position forward to overwrite ep3
                memmove(buf_curr, buf_curr + 1, buf_end - buf_curr);

                // buf_curr unchanged
                buf_end--;
                remove_bytes++;
            }

            zeros = 0;
        }

        if (!*buf_curr)
        {
            zeros++;
        } else
        {
            zeros = 0;
        }
    }

    ni_log(NI_LOG_TRACE, "%s: %d, exit\n", __func__, remove_bytes);
    return remove_bytes;
}

/*!******************************************************************************
*  \brief  overwrite the 32 bits of integer value at bit position pos
*
*  \param  buf   data buffer to be worked on
*          pos   the position to be modified
*          value The value that needs to be modified to
*
*  \return void
*
* Note: caller *MUST* ensure that the pos and value won't go beyond the memory
*        boundary of data. otherwise memory corruption would occur.
******************************************************************************/
void ni_logan_overwrite_specified_pos(uint8_t *buf, int pos, int value)
{
  int pos_byte    = (pos/8);
  int pos_in_byte = pos%8;
  int remaining_bytes_in_current_byte = 8 - pos_in_byte;

  if (pos_in_byte == 0) // at beginning of the byte
  {
    buf[pos_byte] = (uint8_t)(value >> 24);
    buf[pos_byte+1] = (uint8_t)(value >> 16);
    buf[pos_byte+2] = (uint8_t)(value >> 8);
    buf[pos_byte+3] = (uint8_t)(value);
  }
  else
  {
    buf[pos_byte]   = buf[pos_byte] + (uint8_t)(value >> (32-remaining_bytes_in_current_byte));
    buf[pos_byte+1] = (uint8_t)(value >> (32-remaining_bytes_in_current_byte-8));
    buf[pos_byte+2] = (uint8_t)(value >> (32-remaining_bytes_in_current_byte-16));
    buf[pos_byte+3] = (uint8_t)(value >> (32-remaining_bytes_in_current_byte-24));
    buf[pos_byte+4] = buf[pos_byte+4] + ((uint8_t)(value << remaining_bytes_in_current_byte));
  }
}

#ifdef MEASURE_LATENCY
/*!******************************************************************************
*  \brief  Create a latency measurement queue object of a given capacity
*
*  \param  capacity maximum size of queue
*
*  \return ni_logan_lat_meas_q_t latency measurement queue structure
*
*******************************************************************************/
ni_logan_lat_meas_q_t * ni_logan_lat_meas_q_create(unsigned capacity)
{
  ni_logan_lat_meas_q_t* queue = (ni_logan_lat_meas_q_t*) malloc(sizeof(ni_logan_lat_meas_q_t));
  if (!queue)
  {
    ni_log(NI_LOG_ERROR, "ERROR %d: Failed to allocate memory for lat_meas-queue queue",
           NI_ERRNO);
    return NULL;
  }
  queue->capacity = capacity;
  queue->front = queue->size = 0;
  queue->rear = capacity - 1;
  queue->array = (ni_logan_lat_meas_q_entry_t*) malloc(queue->capacity * sizeof(ni_logan_lat_meas_q_entry_t));
  if(!queue->array)
  {
    ni_log(NI_LOG_ERROR, "ERROR %d: Failed to allocate memory for lat_meas_queue queue->array",
           NI_ERRNO);
    return NULL;
  }
  return queue;
}

/*!******************************************************************************
*  \brief  Push an item onto the queue
*
*  \param  queue pointer to latency queue
*  \param  item ni_logan_lat_meas_q_entry_t item to push onto the queue
*
*  \return void 1 if success, NULL if failed
*
*******************************************************************************/
void * ni_logan_lat_meas_q_enqueue(ni_logan_lat_meas_q_t* queue, ni_logan_lat_meas_q_entry_t item)
{
  if (queue->size == queue->capacity)
  {
    return NULL;
  }
  queue->rear = (queue->rear + 1) % queue->capacity;
  queue->array[queue->rear] = item;
  queue->size = queue->size + 1;
  return (void *) 1;
}

/*!******************************************************************************
*  \brief  Pop an item from the queue
*
*  \param  queue pointer to latency queue
*
*  \return void pointer to popped item
*
*******************************************************************************/
void * ni_logan_lat_meas_q_dequeue(ni_logan_lat_meas_q_t* queue)
{
  ni_logan_lat_meas_q_entry_t *dequeue_item;
  if (queue->size == 0)
  {
    return NULL;
  }
  dequeue_item = &queue->array[queue->front];
  queue->front = (queue->front + 1)%queue->capacity;
  queue->size = queue->size - 1;
  return dequeue_item;
}

/*!******************************************************************************
*  \brief  Get a pointer to rear of queue
*
*  \param  queue pointer to latency queue
*
*  \return void pointer to rear of queue
*
*******************************************************************************/
void * ni_logan_lat_meas_q_rear(ni_logan_lat_meas_q_t* queue)
{
  return queue->size == 0 ? NULL : &queue->array[queue->rear];
}

/*!******************************************************************************
*  \brief  Get a pointer to front of queue
*
*  \param  queue pointer to latency queue
*
*  \return void pointer to front of queue
*
*******************************************************************************/
void * ni_logan_lat_meas_q_front(ni_logan_lat_meas_q_t* queue)
{
  return queue->size == 0 ? NULL : &queue->array[queue->front];
}

/*!******************************************************************************
*  \brief  Add a new entry to latency queue
*
*  \param  dec_frame_time_q pointer to latency queue
*  \param  abs_time frame start time for latency comparison
*  \param  ts_time reference frame timestamp time
*
*  \return void 1 if success, NULL if failed
*
*******************************************************************************/
void * ni_logan_lat_meas_q_add_entry(ni_logan_lat_meas_q_t *dec_frame_time_q,
                               uint64_t abs_time,
                               int64_t ts_time)
{
  // ni_log(NI_LOG_INFO, "ni_logan_lat_meas_q_add_entry abs_time=%lu ts_time=%ld\n");
  ni_logan_lat_meas_q_entry_t entry = {.abs_timenano = abs_time, .ts_time = ts_time};
  return ni_logan_lat_meas_q_enqueue(dec_frame_time_q, entry);
}

/*!******************************************************************************
*  \brief  Check latency of a frame referenced by its timestamp
*
*  \param  dec_frame_time_q pointer to latency queue
*  \param  abs_time frame end time for latency comparison
*  \param  ts_time reference frame timestamp time
*
*  \return uint64_t value of latency if suceeded, -1 if failed
*
*******************************************************************************/
uint64_t ni_logan_lat_meas_q_check_latency(ni_logan_lat_meas_q_t *dec_frame_time_q,
                                     uint64_t abs_time,
                                     int64_t ts_time)
{
  // ni_log(NI_LOG_INFO, "ni_logan_lat_meas_q_check_latency abs_time=%lu ts_time=%ld\n");
  uint32_t dequeue_count = 0;
  ni_logan_lat_meas_q_entry_t * entry = ni_logan_lat_meas_q_front(dec_frame_time_q);

  if (entry == NULL)
  {
    return -1;
  }

  if (entry->ts_time == ts_time)
  {
    ni_logan_lat_meas_q_dequeue(dec_frame_time_q);
    dequeue_count++;
  }
  else
  {
    while (entry->ts_time < ts_time)
    {
      // queue miss, perhaps frame was not decoded properly or TS was offset
      entry = ni_logan_lat_meas_q_dequeue(dec_frame_time_q);
      dequeue_count++;
      if (entry == NULL)
      {
        return -1;
      }
    }
  }
  // ni_log(NI_LOG_INFO, "DQ_CNT:%d,QD:%d,", dequeue_count, dec_frame_time_q->size);

  if (entry == NULL)
  {
    // queue overrun
    return -1;
  }
  else if (entry->ts_time > ts_time)
  {
    // queue miss, perhaps frame was not enqueued properly or TS was offset
    return -1;
  }
  else if (entry->ts_time == ts_time)
  {
    // queue item is perfectly matched, calculate latency
    return (abs_time - entry->abs_timenano);
  }
}
#endif
