#!/usr/bin/env bash

IDX_A=0
IDX_B=0
XCODS=1
NO_NUMA=0
DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" >/dev/null 2>&1 && pwd )"
STIME=$SECONDS
declare -a FFMPEG_PIDS

# check if $1 is in array $2
function containsElement() {
    local e match="$1"
    shift
    for e; do [[ "$e" == "$match" ]] && return 1; done
    return 0
}

function show_help() {
    echo "Usage: ./amd_multi_inst_demo.sh [OPTION]"
    echo "Start a number of FFmpeg transcoding processes with NUMA node pinning for AMD ZEN1"
    echo "Note, input test clip must be h264 for this demo"
    echo "FFmpeg logs are stored in ./output_logs/"
    echo "Transcoded videos are stored in ./output_vids/"
    echo "Example: ./amd_multi_inst_demo.sh -a 0 -b 3 -x 4 -i '~/input_test_clip.mp4'"
    echo ""
    echo "Options:"
    echo "-h, --help"
    echo "    display this help and exit"
    echo "-a START_IDX"
    echo "    index of first T408 card in range to test. (Default: 0)"
    echo "-b END_IDX"
    echo "    index of last T408 card in range to test. (Default: 0)"
    echo "-x INSTANCES_PER_CARD"
    echo "    number of transcoding jobs to run per card. (Default: 1)"
    echo "--no_numa_pinning"
    echo "    disable numa node pinning to see bad performance."
    echo "-i INPUT_VIDEO"
    echo "    path to H264 input test clip to use."
}

# parse a flag with an arg in or after it
# $1 flag pattern, $2 entire flag arg, $3 arg after flag arg
# return 1 if path is in second arg (separated by space), else return 0. Store path in $extract_arg_ret
function extract_arg() {
    unset extract_arg_ret
    # check valid arg flag
    if [ -n "$(printf "%s" ${2} | grep -Eoh "${1}")" ]; then
        # check if path string is connected by '=' or is in following arg
        if [ -n "$(echo "${2}" | grep -Eoh "${1}=")" ]; then
            arg_str=`printf "%s" "${2}" | grep -Poh "${1}=\K.+"`;
            # trim out leading and trailing quotation marks
            extract_arg_ret=`echo "${arg_str}" | sed -e 's/^\(["'\'']\)//' -e 's/\(["'\'']\)$//'`;
            return 0;
        elif [ -n "$(printf "%s" ${2} | grep -Eoh "^${1}$")" ]; then
            arg_str="${3}";
            # trim out leading and trailing quotation marks
            extract_arg_ret=`printf "%s" "${arg_str}" | sed -e 's/^\(["'\'']\)//' -e 's/\(["'\'']\)$//'`;
            return 1;
        else
            echo "Unknown option '$2', exiting";
            exit 1;
        fi
    else
        echo "Target flag '$1' not found in '$2', exiting"; exit 1;
    fi
}

function parse_user_options() {
    while [ "$1" != "" ]; do
        case $1 in
            -h | --help)        show_help; exit 0;;
            -a | -a=* )         extract_arg "\-a" $1 $2; prev_rc=$?;
                                if [ "$prev_rc" -eq 1 ]; then shift; fi
                                IDX_A=$extract_arg_ret;;
            -b | -b=* )         extract_arg "\-b" $1 $2; prev_rc=$?;
                                if [ "$prev_rc" -eq 1 ]; then shift; fi
                                IDX_B=$extract_arg_ret;;
            -x | -x=* )         extract_arg "\-x" $1 $2; prev_rc=$?;
                                if [ "$prev_rc" -eq 1 ]; then shift; fi
                                XCODS=$extract_arg_ret;;
            --no_numa_pinning)  NO_NUMA=1;;
            -i | -i=*)          extract_arg "\-i" $1 $2; prev_rc=$?;
                                if [ "$prev_rc" -eq 1 ]; then shift; fi
                                INPUT_VID=$extract_arg_ret;;
            *)                  echo "Unknown option $1, exiting"; exit 1;;
        esac
        shift
    done

    # Check input args
    if [[ $XCODS = 0 ]]; then
        echo "BAD ARGS: No jobs set. Exiting"; exit 1;
    fi
    if [ "${XCODS}" -gt 32 ]; then
        echo "BAD ARGS: No jobs set. Exiting"; exit 1;
    fi
    if [ -z ${INPUT_VID+x} ]; then
        echo "BAD ARGS: No input video set. Exiting"; exit 1;
    fi
    if [ "${IDX_A}" -gt "${IDX_B}" ]; then
        echo "BAD ARGS: Card start index and end index do not for valid range. Exiting"; exit 1;
    fi

    # Check selected range of cards exist
    card_idx_arr=( $(ni_rsrc_list_logan 2>&1 | grep -Poh "(?<=Decoder #)[0-9]+") )
    for i in $(seq $IDX_A $IDX_B); do
        if containsElement $i "${card_idx_arr[@]}"; then
            echo "BAD ARGS: selected range of cards to use include ${i} which is not detected in system"; exit 1;
        fi
    done
}

function spinner() {
    i=1
    sp="/-\|"
    while true
    do
        CTIME=$SECONDS
        ETIME=$(( CTIME - STIME ))
        TOTAL_FPS=0
        TOTAL_STREAMS=0

        for log_file in ${DIR}/output_logs/ffmpeg_log-*-*.txt
        do
            iFPS=`tac $log_file | grep -m 1 -sPoh 'fps=[ ]*\K[0-9\.]+' | tail -n 1`
            if [[ $iFPS == "0.0" ]] || [[ $iFPS == "" ]]; then continue; fi
            TOTAL_FPS=$(awk "BEGIN {print $TOTAL_FPS + $iFPS}")
            TOTAL_STREAMS=$((TOTAL_STREAMS + 1))
        done

        printf "\r${sp:i++%${#sp}:1} [${ETIME}s] total streams: %3d; average fps: %6s" "${TOTAL_STREAMS}" "$(bc <<< "scale=2; ${TOTAL_FPS} / ${TOTAL_STREAMS}")"
        sleep 0.5
    done
}
      
function end_test() {
    CTIME=$SECONDS
    ETIME=$(( CTIME - STIME ))

    # sudo killall ffmpeg
    for pid in "${FFMPEG_PIDS[@]}"; do
        kill -s SIGTERM "${pid}" &> /dev/null
    done
    kill $SPINNER_PID

    printf "\nStopped amd_multi_inst_demo. Elapsed time: ${ETIME}s\n"
    trap - EXIT
    exit 0
}

# -------------------------------- MAIN --------------------------------
parse_user_options $@
echo "Copyright (C) 2019 NETINT Technologies. All rights reserved."
echo "Starting NETINT T408 AMD MULTI INST DEMO"

# initialize T408 resources, set CPU to performance mode, configure output folders
echo "Running setup..."
ni_rsrc_mon_logan &> /dev/null
for i in {0..127}; do sudo sh -c "echo performance > /sys/devices/system/cpu/cpu${i}/cpufreq/scaling_governor"; done 2> /dev/null
sudo rm -rf ./output_logs/ &> /dev/null
mkdir ./output_logs/ &> /dev/null
sudo rm -rf ./output_vids/ &> /dev/null
mkdir ./output_vids/ &> /dev/null

# get NUMA associations, print FFmpeg commands, run FFmpeg commands
echo "Running $(( IDX_B - IDX_A + 1 )) x ${XCODS} transcodes of ${INPUT_VID}"
echo "To terminate test use ctrl^C or terminate process ID: $$"
trap end_test EXIT

numa_csv=`./t408_numa_associations.sh --csv`

echo "FFmpeg commands used:"
echo "--------------------------------------------------------------------------------"
for i in $(seq $IDX_A $IDX_B); do
    numa_node=`echo ${numa_csv} | grep -Poh "(?<=${i},${i},)[0-9]+"`
    for j in $(seq 1 $XCODS); do
        if [[ $NO_NUMA != 1 ]]; then
            cmd_str="sudo numactl --cpubind=${numa_node} --membind=${numa_node} \
                     ffmpeg -nostdin -ni_interval_fps 1 -vsync 0 \
                     -c:v h264_ni_logan_dec -dec ${i} -i ${INPUT_VID} \
                     -c:v h265_ni_logan_enc -enc ${i} \
                     -xcoder-params intraPeriod=120:RcEnable=1:bitrate=4000000:repeatHeaders=1 \
                     -f mp4 output_vids/t408_xcod_vid_h265_4M-${i}-${j}.mp4"
        else
            cmd_str="ffmpeg -nostdin -ni_interval_fps 1 -vsync 0 \
                     -c:v h264_ni_logan_dec -dec ${i} -i ${INPUT_VID} \
                     -c:v h265_ni_logan_enc -enc ${i} \
                     -xcoder-params intraPeriod=120:RcEnable=1:bitrate=4000000:repeatHeaders=1 \
                     -f mp4 output_vids/t408_xcod_vid_h265_4M-${i}-${j}.mp4"
        fi
        echo $cmd_str
        echo $cmd_str > output_logs/ffmpeg_log-$i-$j.txt
        echo "--------------------------------------------------------------------------------" >> output_logs/ffmpeg_log-$i-$j.txt
        $cmd_str &>> output_logs/ffmpeg_log-$i-$j.txt & FFMPEG_PIDS+=( $! )
        echo "--------------------------------------------------------------------------------"
        sleep 0.1
    done
    unset numa_node
done

# start performance monitor, wait for all jobs to complete
spinner 2> /dev/null &
SPINNER_PID=$!

# sleep until jobs are done.
for pid in "${FFMPEG_PIDS[@]}"; do
   wait "${pid}"
   printf "\nFFmpeg transcoding job ${pid} has completed\n"
done
end_test
