#!/usr/bin/env bash

function parse_user_options() {
    while [ "$1" != "" ]; do
        case $1 in
            -h | --help)            show_help; exit 0;;
            -v | --verbose)         VERBOSE=YES;;
            -d | --install_deps)    sudo apt-get install hwloc numactl; exit 0;;
            -m |-c | --csv)         SYSTEM=YES;;
        esac
        shift
    done
}

function show_help() {
    echo "Usage: ./t408_numa_associations.sh [OPTIONS]"
    echo "Report the dec/enc index to numa node mapping."
    echo "Note, to use NUMA node pinning the dependencies hwloc, and numactl must be installed."
    echo ""
    echo "Options:"
    echo "-h, --help"
    echo "    Print this help"
    echo "-v, --verbose"
    echo "    Verbose output"
    echo "-d, --install_deps"
    echo "    Install dependencies needed for the script, then exit"
    echo "-m, -c, --csv"
    echo "    Output is produced in the form of a csv array with each line containing "
    echo "    information regarding one card. The format of each line return is:"
    echo "    [decoder#][encoder#][numa node#]"
    echo "    Ex. for a system with 2 T408 cards on 2 numa nodes (50 and 52):"
    echo "        0,0,50"
    echo "        1,1,52"
}

# -------------------------------- MAIN --------------------------------
parse_user_options $@

nilspci=`sudo lspci -vn -d "1d82:0202" | grep -A3 "1d82:0202"`
TOTAL_CARDS=`echo ${nilspci} | grep -o "1d82:0202" | wc -l`
if (( `echo ${nilspci} | grep -o "NUMA node" | wc -l` < $TOTAL_CARDS )); then
    echo "ERROR: lspci does not report NUMA nodes for this system. Perhaps dependencies are not installed?"
    exit 1
fi

#declare arrays to hold data we need
declare -a NVME_DEVS
declare -a T408_ENCS
declare -a T408_DECS

BUS_IDS=( $(echo ${nilspci} | grep -Poh "[a-f0-9]{2}:[0-9]{2}\.[0-9](?= 0108:)") )
SLOTS=( $(echo ${nilspci} | grep -Poh "(?<=Physical Slot: )[0-9-]+") )
NUMA_NODES=( $(echo ${nilspci} | grep -Poh "(?<=, NUMA node )[0-9]{1,3}") )
for bus_id in "${BUS_IDS[@]}"; do
    NVME_DEV=`grep "${bus_id}" /sys/class/nvme/*/device/uevent`
    IFS='/' TOKENS=( $NVME_DEV )
    NVME_DEVS+=( ${TOKENS[4]} )
    T408_ENCS+=( $(ni_rsrc_list_logan 2>&1 | grep -B1 -w ${TOKENS[4]} | grep -Poh "(?<=Encoder #)\d{1,3}") )
    T408_DECS+=( $(ni_rsrc_list_logan 2>&1 | grep -B1 -w ${TOKENS[4]} | grep -Poh "(?<=Decoder #)\d{1,3}") )
done

#print the report
if [[ $SYSTEM == "YES" ]]; then
    declare -a RETURN
    for (( i=0; i<TOTAL_CARDS; i++ )); do
        RETURN+=( $(printf "%d,%d,%d\n" "${T408_DECS[$i]}" "${T408_ENCS[$i]}" "${NUMA_NODES[$i]}") )
    done
    printf "%s\n" "${RETURN[@]}"
else
    printf "Total T-408 devices in the system: ${TOTAL_CARDS}\n"
    if [[ $VERBOSE == "YES" ]]; then
        printf "+-------------+-------------+-------------+---------------+-----------------+---------------+\n"
        printf "|  DEC index  |  ENC index  |  NUMA node  |  PCIe Bus ID  |  Physical Slot  |  NVME Device  |\n"
        printf "+-------------+-------------+-------------+---------------+-----------------+---------------+\n"
    else
        printf "+-------------+-------------+-------------+\n"
        printf "|  DEC index  |  ENC index  |  NUMA node  |\n"
        printf "+-------------+-------------+-------------+\n"
    fi

    for (( i=0; i < TOTAL_CARDS; i++ )); do
        if [[ $VERBOSE == "YES" ]]; then
            printf "|  %9d  |  %9d  |  %9d  |  %11s  |  %13s  |  %11s  |\n" "${T408_DECS[$i]}" "${T408_ENCS[$i]}" "${NUMA_NODES[$i]}" "${BUS_IDS[$i]}" "${SLOTS[$i]}" "/dev/${NVME_DEVS[$i]}"
        else
            printf "|  %9d  |  %9d  |  %9d  |\n" "${T408_DECS[$i]}" "${T408_ENCS[$i]}" "${NUMA_NODES[$i]}"
        fi

        if [ $i -eq $((TOTAL_CARDS-1)) ]; then
            if [[ $VERBOSE == "YES" ]]; then
                printf "+-------------+-------------+-------------+---------------+-----------------+---------------+\n"
            else
                printf "+-------------+-------------+-------------+\n"
            fi
            # print the example ffmpeg command with numactl pinning
            printf "Example FFMpeg command for H264->H265 transcoding on dec/enc ${T408_DECS[$i]} with numa node pinning to numa node ${NUMA_NODES[$i]}:\n"
            printf "    sudo numactl --cpubind=${NUMA_NODES[$i]} --membind=${NUMA_NODES[$i]} ffmpeg -c:v h264_ni_logan_dec -dec %s -i <INPUT H264 FILE PATH> -c:v h265_ni_logan_enc -enc %s -f <OUTPUT H265 FILE PATH>\n" "${T408_DECS[$i]}" "${T408_ENCS[$i]}"
        fi
    done
fi #if SYSTEM
