#!/usr/bin/env python3

"""Healthy Scan Tool
Copyright (c) 2019 NETINT Technologies Inc. All rights reserved.
"""

import os
import sys

import argparse
import ast
import itertools
import logging
import re
import subprocess
import time

try:
    from subprocess import DEVNULL
except ImportError:
    import os
    DEVNULL = open(os.devnull, 'wb')

class T408():

    counter = 0

    def __init__(self, serial_number, device_path, fw_rev):
        self.sn = serial_number
        self.device_path = device_path
        self.fw_rev = fw_rev
        self.num_job_instance = 0
        self.realtime_load = 0
        self.card_index = None
        self.result = -1
        self.process = None

    @property
    def sn(self):
        return self._sn
    @sn.setter
    def sn(self, value):
        match = re.search(r"^T[AU]\d{2}-\d{2}-\d{2}-[A-Z]\d{2}-\d{3,4}[A-Z]?$", value.upper())
        if match:
            self._sn = match.group()
        else:
            raise TypeError("%s is not a valid Serial Number" % value)

    @property
    def device_path(self):
        return self._device_path
    @device_path.setter
    def device_path(self, value):
        match = re.search(r"/dev/nvme\d+", value)
        if match:
            self._device_path = match.group()
        else:
            raise TypeError("%s is not a valid nvme device_path" % value)

    @property
    def fw_rev(self):
        return self._fw_rev
    @fw_rev.setter
    def fw_rev(self, value):
        match = re.search(r"^[A-Za-z0-9]{3}", value)
        if match:
            if (int(match.group()[0])==1 and int(match.group())>=151) or (int(match.group()[0])>1):
                self._fw_rev = match.group()
                return
        raise TypeError("%s is not a valid Firmware Revision" % value)

    @property
    def result(self):
        return self._result
    @result.setter
    def result(self, value):
        """ Set new result value
        -1:Under testing
        0: Pass
        1: Will not test. Existing load/instance
        2: Error: Cannot find device in ni_rsrc_mon_logan
        3: Error: Cannot find an instance after starting ffmpeg process
        4: Error: Cannot find load after starting ffmpeg process
        5: Error: Instance value incorrect after stopping ffmpeg process
        6: Error: Load value incorrect after stopping ffmpeg process
        """
        if value < -1:
            raise ValueError("%s is not a valid result code" % value)
        if value == 1:
            logging.debug("[WARNING] : %s %s has an existing load.  Removing from test"
                    % (self.sn, self.device_path))
        elif value == 2:
            logging.debug("[ERROR] : %s %s is not found in ni_rsrc_mon_logan"
                    % (self.sn, self.device_path))
        elif value == 3:
            logging.debug("[ERROR] : %s %s cannot find an instance after starting ffmpeg process"
                    % (self.sn, self.device_path))
        elif value == 4:
            logging.debug("[ERROR] : %s %s cannot find a load after starting ffmpeg process"
                    % (self.sn, self.device_path))
        elif value == 5:
            logging.debug("[ERROR] : %s %s instance is not 0 after stopping ffmpeg process"
                    % (self.sn, self.device_path))
        elif value == 6:
            logging.debug("[ERROR] : %s %s load is not 0 after stopping ffmpeg process"
                    % (self.sn, self.device_path))
        self._result = value

    def run_ffmpeg(self, cmd):
        if self.result == -1:
            T408.counter += 1
            time.sleep(0.1)
            self.process = subprocess.Popen(cmd.format(self.card_index).split(),
                    stdout=DEVNULL, stderr=subprocess.STDOUT)


def arg_parse():
    parser = argparse.ArgumentParser(description='Healthy Scan Tool')
    parser.add_argument('--max_runtime',
                        type=int,
                        default = 180,
                        help='Test will force stop after max_runtime. Default, 3min')
    parser.add_argument('--seed_path',
                        type=str,
                        default = "./test_clip.h264.list",
                        help='Seed video. Default: test_clip.h264.list'
                        )
    parser.add_argument('--serial_path',
                        type=str,
                        default = "./T408_serial_numbers.txt",
                        help='Text file of serial numbers. Default: T408_serial_numbers.txt'
                        )
    parser.add_argument('--verbose',
                        action='store_true',
                        help="Set verbose logging"
                        )
    parser.add_argument('--yuvbypass',
                        action='store_true',
                        help="Enable yuvbypass (hwframe 1)"
                        )
    return parser.parse_args()

def create_ffmpeg_cmd(source_file, yuvbypass=False):
    """Generate ffmpeg testing command based on source file"""
    if not os.path.isfile(source_file):
        raise FileNotFoundError("No such file or directory: '%s'" % source_file)
    status, _ = subprocess.getstatusoutput("ffprobe %s" % source_file)
    if yuvbypass == True:
        cmd = "sudo /usr/local/bin/ffmpeg -hide_banner -vsync 0 -hwframe 1 -nostdin "
    else:
        cmd = "sudo /usr/local/bin/ffmpeg -hide_banner -vsync 0 -nostdin "
    if status == 0:
        cmd += "-c:v h264_ni_logan_dec -re -dec 0 -i %s " % source_file
    else:
        cmd += "-c:v h264_ni_logan_dec -f concat -re -dec {0} -i %s " % source_file
    cmd += "-c:v h264_ni_logan_enc -enc {0} -f null /dev/null"
    return cmd

def get_nvme_list():
    """Return a dict serial,path pairs from 'sudo nvme list'"""
    command = "sudo nvme list -o json"
    logging.debug("Parsing '%s' for Serial Numbers:" % command)
    status, output = subprocess.getstatusoutput(command)
    if status != 0:
        raise Exception("Command '%s' failed " % command)
    output = ast.literal_eval(output)
    nvme_list = {}
    for i in output["Devices"]:
        nvme_list[i["SerialNumber"]] = [i["DevicePath"], i["Firmware"]]
        logging.debug("%s : %s" % (i["SerialNumber"],i["DevicePath"]))
    if len(nvme_list) < 1:
        raise Exception("Cannot find any cards via '%s'" % command)
    return nvme_list

def get_text_serial(f_serial, nvme_list):
    """Return a list of serial numbers from the provided text"""
    logging.debug("Parsing '%s' for T408 Serial Numbers:" % f_serial)
    t408_text_serial = []
    t408_text_not_in_nvme_list = []
    if not os.path.isfile(f_serial):
        raise Exception("Cannot find `%s`. Please run `python3 %s --help`"
                % (f_serial,os.path.basename(__file__)))

    with open(f_serial, 'r') as f:
        for line in f:
            line = line.strip()
            if line.startswith("#") or line.startswith("//"):
                continue
            for serial in line.split():
                try:
                    t408_text_serial.append(T408(serial, nvme_list[serial][0], nvme_list[serial][1]))
                    logging.debug(serial)
                except TypeError:
                    logging.debug("[WARNING] : Unknown Serial Number in %s - %s -- skipping"
                          % (f_serial,serial))
                except KeyError:
                    t408_text_not_in_nvme_list.append(serial)
                    logging.debug("[WARNING] : %s does not exist in 'nvme list' -- skipping"
                          % serial)
    if len(t408_text_serial) < 1:
        raise Exception("Cannot find any T408 Serial Number from %s in 'nvme list'. "
                        "Please double check %s" % (f_serial, f_serial))
    return t408_text_serial,t408_text_not_in_nvme_list

def print_t408(t408):
    """Print all attributes"""
    logging.debug("")
    for i in t408:
        logging.debug(i.__dict__)


def sleep_spinner(t=10):
    """replaces time.sleep(); sleep with a spinner display"""
    spinner = itertools.cycle(['-', '/', '|', '\\'])
    timer = time.time()
    while time.time()-timer < t:
        sys.stdout.write(next(spinner))
        sys.stdout.flush()
        sys.stdout.write('\b')
        time.sleep(0.5)

def update_ni_rsrc_mon(t408):
    """Return a dict of ni_rsrc_mon_logan information"""
    command = "ni_rsrc_mon_logan"
    status, output = subprocess.getstatusoutput(command)
    if status != 0:
        raise Exception("Command '%s' failed " % command)
    ni_rsrc_mon_logan = {'decoders':{},'encoders':{}}
    match = re.findall(r'\n[\s|\w]\s+(\d+)\s+(\d+)\s+(\d+)\s+\d+\s+(\d+)\s+/dev/nvme\d+\s+(/dev/nvme\d+)n\d?', output)
    if len(match) < 1:
        raise Exception("Cannot find any cards via '%s'" % command)
    for i in t408:
        if i.result != -1:
            continue
        for data in match:
            if i.device_path == data[4]:
                i.card_index = int(data[0])
                i.realtime_load = int(data[1])
                i.num_job_instance = int(data[3])
                break
        else:
            i.result = 2
    #print_t408(t408)

def main():

    args = arg_parse()

    print("Please remember to update the serial number text file. eg ./T408_serial_numbers.txt")
    prompt = input("Continue? [Y/N]").lower()
    if prompt != 'y' and prompt != "yes":
        print("exiting...")
        return 2

    if args.verbose == True:
        logging.basicConfig(format='%(asctime)s %(message)s',
                            datefmt='%m/%d/%Y %I:%M:%S %p',
                            level=logging.DEBUG)

    # Gather all card information and rsrc mon
    nvme_list = get_nvme_list()
    t408,sn_not_in_nvme_list = get_text_serial(args.serial_path, nvme_list)
    update_ni_rsrc_mon(t408)

    # check existing loads and instances
    for i in t408:
        if i.num_job_instance > 0 or i.realtime_load > 0:
            i.result = 1
            print("Found existing traffic on %s; skipping test on this card" % i.sn)

    # start ffmpeg test
    base_cmd = create_ffmpeg_cmd(args.seed_path, args.yuvbypass)
    logging.critical("Starting Test...")
    logging.critical(base_cmd)
    for i in t408:
        i.run_ffmpeg(base_cmd)

    # check new loads and instances
    start_time = time.time()
    counter = 0
    while time.time()-start_time < args.max_runtime:
        sleep_spinner(10)
        update_ni_rsrc_mon(t408)
        for i in t408:
            if i.result != -1:
                continue
            if i.process.poll() == None:
                # ffmpeg process still ongoing
                if i.num_job_instance != 1 or i.realtime_load < 1:
                    # recheck
                    time.sleep(2)
                    update_ni_rsrc_mon(t408)
                    if i.process.poll() == None:
                        if i.num_job_instance != 1:
                            i.result = 3
                        elif i.realtime_load < 1:
                            i.result = 4
            else:
                # ffmpeg process stopped
                counter += 1
                if i.num_job_instance != 0 or i.realtime_load > 0:
                    # recheck
                    time.sleep(2)
                    update_ni_rsrc_mon(t408)
                    if i.num_job_instance != 0:
                        i.result = 5
                    elif i.realtime_load > 0:
                        i.result = 6
                    else:
                        i.result = 0
                else:
                    i.result = 0
        if T408.counter == counter:
            break

    #print results
    print("1. T408 cards - good status")
    for i in t408:
        if i.result == 0:
            print(i.sn)
    print("2. T408 cards - bad status")
    for i in t408:
        if i.result > 1:
            print(i.sn)
    for i in sn_not_in_nvme_list:
        print(i)
    if args.verbose == True:
        print("3. T408 cards - untested; Found existing traffic")
        for i in t408:
            if i.result == 1:
                print(i.sn)
        print("4. T408 cards - Not found in 'nvme list'")
        for i in sn_not_in_nvme_list:
            print(i)

if __name__ == '__main__':
    main()
